@@ -4,12 +4,12 @@
 # shellcheck source=contrib/jenkins_common.sh
 . $(dirname "$0")/jenkins_common.sh
 
-osmo-build-dep.sh libosmocore "" --disable-doxygen
-
 export PKG_CONFIG_PATH="$inst/lib/pkgconfig:$PKG_CONFIG_PATH"
 export LD_LIBRARY_PATH="$inst/lib"
 
-osmo-build-dep.sh libosmo-abis
+osmo-build-dep.sh libosmocore "" --disable-doxygen
+osmo-build-dep.sh libosmo-netif "" --disable-doxygen
+osmo-build-dep.sh libosmo-abis "" --disable-dahdi
 
 cd "$deps"
 osmo-layer1-headers.sh oct "$FIRMWARE_VERSION"
