@@ -5,13 +5,6 @@
 
 #include <osmo-bts/gsm_data.h>
 
-/* Whether a logical channel must be activated automatically */
-#define TRX_CHAN_FLAG_AUTO_ACTIVE	(1 << 0)
-
-/* FIXME: we should actually activate 'auto-active' channels */
-#define TRX_CHAN_IS_ACTIVE(state, chan) \
-	(trx_chan_desc[chan].flags & TRX_CHAN_FLAG_AUTO_ACTIVE || (state)->active)
-
 #define TRX_GMSK_NB_TSC(br) \
 	_sched_train_seq_gmsk_nb[(br)->tsc_set][(br)->tsc]
 
@@ -97,10 +90,10 @@ struct l1sched_chan_state {
 	bool			active;		/* Channel is active */
 	ubit_t			*dl_bursts;	/* burst buffer for TX */
 	enum trx_mod_type	dl_mod_type;	/* Downlink modulation type */
+	uint8_t			dl_mask;	/* mask of transmitted bursts */
 	sbit_t			*ul_bursts;	/* burst buffer for RX */
-	sbit_t			*ul_bursts_prev;/* previous burst buffer for RX (repeated SACCH) */
 	uint32_t		ul_first_fn;	/* fn of first burst */
-	uint8_t			ul_mask;	/* mask of received bursts */
+	uint32_t		ul_mask;	/* mask of received bursts */
 
 	/* loss detection */
 	uint32_t		last_tdma_fn;	/* last processed TDMA frame number */
@@ -138,7 +131,7 @@ struct l1sched_chan_state {
 	/* Uplink measurements */
 	struct {
 		/* Active channel measurements (simple ring buffer) */
-		struct l1sched_meas_set buf[8]; /* up to 8 entries */
+		struct l1sched_meas_set buf[24]; /* up to 24 (BUFMAX) entries */
 		unsigned int current; /* current position */
 
 		/* Interference measurements */
@@ -192,6 +185,12 @@ int trx_sched_set_pchan(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config pch
 /*! \brief set all matching logical channels active/inactive */
 int trx_sched_set_lchan(struct gsm_lchan *lchan, uint8_t chan_nr, uint8_t link_id, bool active);
 
+/*! \brief set uplink access on given logical channels active/inactive */
+int trx_sched_set_ul_access(struct gsm_lchan *lchan, uint8_t chan_nr, bool active);
+
+/*! \brief set all logical channels of BCCH/CCCH active/inactive */
+int trx_sched_set_bcch_ccch(struct gsm_lchan *lchan, bool active);
+
 /*! \brief set mode of all matching logical channels to given mode(s) */
 int trx_sched_set_mode(struct gsm_bts_trx_ts *ts, uint8_t chan_nr, uint8_t rsl_cmode,
 	uint8_t tch_mode, int codecs, uint8_t codec0, uint8_t codec1,
@@ -304,6 +303,10 @@ int trx_sched_ul_burst(struct l1sched_ts *l1ts, struct trx_ul_burst_ind *bi);
 
 /* Averaging mode for trx_sched_meas_avg() */
 enum sched_meas_avg_mode {
+	/* first 22 of last 24 bursts (for TCH/F14.4, TCH/F9.6, TCH/F4.8) */
+	SCHED_MEAS_AVG_M_S24N22,
+	/* last 22 bursts (for TCH/H4.8, TCH/H2.4) */
+	SCHED_MEAS_AVG_M_S22N22,
 	/* last 4 bursts (default for xCCH, PTCCH and PDTCH) */
 	SCHED_MEAS_AVG_M_S4N4,
 	/* last 8 bursts (default for TCH/F and FACCH/F) */
