@@ -21,12 +21,9 @@ extern struct cmd_element cfg_bts_no_auto_band_cmd;
 struct phy_instance *vty_get_phy_instance(struct vty *vty, int phy_nr, int inst_nr);
 
 int bts_vty_go_parent(struct vty *vty);
-int bts_vty_is_config_node(struct vty *vty, int node);
 
 int bts_vty_init(void *ctx);
 
-struct gsm_network *gsmnet_from_vty(struct vty *v);
-
 extern struct vty_app_info bts_vty_info;
 extern struct gsm_bts *g_bts;
 
