@@ -13,7 +13,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -39,10 +39,10 @@
 #include <osmocom/core/signal.h>
 #include <osmocom/core/macaddr.h>
 #include <osmocom/core/fsm.h>
+#include <osmocom/gsm/protocol/ipaccess.h>
+#include <osmocom/gsm/ipa.h>
 #include <osmocom/abis/abis.h>
 #include <osmocom/abis/e1_input.h>
-#include <osmocom/abis/ipaccess.h>
-#include <osmocom/gsm/ipa.h>
 
 #include <osmo-bts/abis.h>
 #include <osmo-bts/logging.h>
@@ -87,10 +87,17 @@ struct abis_link_fsm_priv {
 
 static void reset_oml_link(struct gsm_bts *bts)
 {
+	struct e1inp_sign_link *link;
+
 	if (bts->oml_link) {
 		struct timespec now;
 
-		e1inp_sign_link_destroy(bts->oml_link);
+		/* Mark bts->oml_link ptr null before calling sign_link_destroy,
+		 * to avoid a callback triggering this same code path. */
+		link = bts->oml_link;
+		bts->oml_link = NULL;
+
+		e1inp_sign_link_destroy(link);
 
 		/* Log a special notice if the OML connection was dropped relatively quickly. */
 		if (bts->oml_conn_established_timestamp.tv_sec != 0 && clock_gettime(CLOCK_MONOTONIC, &now) == 0 &&
@@ -100,14 +107,16 @@ static void reset_oml_link(struct gsm_bts *bts)
 			     "A common error is a mismatch between unit_id configuration parameters of BTS and BSC.\n",
 			     (uint64_t) (now.tv_sec - bts->oml_conn_established_timestamp.tv_sec));
 		}
-		bts->oml_link = NULL;
 	}
 	memset(&bts->oml_conn_established_timestamp, 0, sizeof(bts->oml_conn_established_timestamp));
 
 	/* Same for IPAC_PROTO_OSMO on the same ipa connection: */
 	if (bts->osmo_link) {
-		e1inp_sign_link_destroy(bts->osmo_link);
+		/* Mark bts->osmo_link ptr null before calling sign_link_destroy,
+		 * to avoid a callback triggering this same code path. */
+		link = bts->osmo_link;
 		bts->osmo_link = NULL;
+		e1inp_sign_link_destroy(link);
 	}
 
 }
@@ -225,9 +234,12 @@ static void abis_link_connected(struct osmo_fsm_inst *fi, uint32_t event, void *
 
 	/* Then iterate over the RSL signalling links */
 	llist_for_each_entry(trx, &bts->trx_list, list) {
-		if (trx->rsl_link) {
-			e1inp_sign_link_destroy(trx->rsl_link);
-			trx->rsl_link = NULL;
+		if (trx->bb_transc.rsl.link) {
+			/* Mark link ptr null before calling sign_link_destroy,
+			 * to avoid a callback triggering this same code path. */
+			struct e1inp_sign_link *link = trx->bb_transc.rsl.link;
+			trx->bb_transc.rsl.link = NULL;
+			e1inp_sign_link_destroy(link);
 			if (trx == trx->bts->c0)
 				load_timer_stop(trx->bts);
 		} else {
@@ -364,7 +376,7 @@ int abis_bts_rsl_sendmsg(struct msgb *msg)
 
 	/* osmo-bts uses msg->trx internally, but libosmo-abis uses
 	 * the signalling link at msg->dst */
-	msg->dst = msg->trx->rsl_link;
+	msg->dst = msg->trx->bb_transc.rsl.link;
 	return abis_sendmsg(msg);
 }
 
@@ -404,10 +416,10 @@ static struct e1inp_sign_link *sign_link_up(void *unit, struct e1inp_line *line,
 			break;
 		}
 		e1inp_ts_config_sign(sign_ts, line);
-		trx->rsl_link = e1inp_sign_link_create(sign_ts, E1INP_SIGN_RSL,
-						       trx, trx->rsl_tei, 0);
+		trx->bb_transc.rsl.link = e1inp_sign_link_create(sign_ts, E1INP_SIGN_RSL,
+								 trx, trx->bb_transc.rsl.tei, 0);
 		trx_link_estab(trx);
-		return trx->rsl_link;
+		return trx->bb_transc.rsl.link;
 	}
 	return NULL;
 }
@@ -476,8 +488,11 @@ static int inp_s_cbfn(unsigned int subsys, unsigned int signal,
 		return 0;
 
 	struct input_signal_data *isd = signal_data;
-	DEBUGP(DABIS, "Input Signal %s received for link_type=%s\n",
-	       get_value_string(e1inp_signal_names, signal), e1inp_signtype_name(isd->link_type));
+	DEBUGP(DABIS, "Input Signal %s received for ts-%u-%u link_type=%s\n",
+	       get_value_string(e1inp_signal_names, signal),
+	       isd->line ? isd->line->num : -1,
+	       isd->ts_nr,
+	       e1inp_signtype_name(isd->link_type));
 
 	return 0;
 }
