@@ -74,12 +74,17 @@ static void gsm_bts_trx_init_ts(struct gsm_bts_trx *trx)
 		ts->trx = trx;
 		ts->nr = tn;
 
+		ts->tsc_oml_configured = false;
+		ts->tsc_rsl_configured = false;
+		ts->tsc = ts->tsc_oml = ts->tsc_rsl = 0xff;
+
 		ts->mo.fi = osmo_fsm_inst_alloc(&nm_chan_fsm, trx, ts,
 						LOGL_INFO, NULL);
 		osmo_fsm_inst_update_id_f(ts->mo.fi, "%s-ts%u",
 					  trx->bb_transc.mo.fi->id, ts->nr);
 		gsm_mo_init(&ts->mo, trx->bts, NM_OC_CHANNEL,
 			    trx->bts->nr, trx->nr, ts->nr);
+		oml_mo_state_init(&ts->mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
 
 		gsm_bts_trx_ts_init_lchan(ts);
 	}
@@ -99,6 +104,10 @@ void gsm_bts_trx_init_shadow_ts(struct gsm_bts_trx *trx)
 		ts->trx = trx;
 		ts->nr = tn;
 
+		ts->tsc_oml_configured = false;
+		ts->tsc_rsl_configured = false;
+		ts->tsc = ts->tsc_oml = ts->tsc_rsl = 0xff;
+
 		/* Link both primary and shadow */
 		trx->ts[tn].vamos.peer = ts;
 		ts->vamos.peer = &trx->ts[tn];
@@ -148,14 +157,14 @@ struct gsm_bts_trx *gsm_bts_trx_alloc(struct gsm_bts *bts)
 	trx->mo.fi = osmo_fsm_inst_alloc(&nm_rcarrier_fsm, trx, trx,
 						  LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(trx->mo.fi, "bts%d-trx%d", bts->nr, trx->nr);
-	gsm_mo_init(&trx->mo, bts, NM_OC_RADIO_CARRIER,
-		    bts->nr, trx->nr, 0xff);
+	gsm_mo_init(&trx->mo, bts, NM_OC_RADIO_CARRIER, bts->nr, trx->nr, 0xff);
+	oml_mo_state_init(&trx->mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
 
 	trx->bb_transc.mo.fi = osmo_fsm_inst_alloc(&nm_bb_transc_fsm, trx, &trx->bb_transc,
 						  LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(trx->bb_transc.mo.fi, "bts%d-trx%d", bts->nr, trx->nr);
-	gsm_mo_init(&trx->bb_transc.mo, bts, NM_OC_BASEB_TRANSC,
-		    bts->nr, trx->nr, 0xff);
+	gsm_mo_init(&trx->bb_transc.mo, bts, NM_OC_BASEB_TRANSC, bts->nr, trx->nr, 0xff);
+	oml_mo_state_init(&trx->bb_transc.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
 
 	gsm_bts_trx_init_ts(trx);
 
@@ -222,15 +231,18 @@ int trx_link_estab(struct gsm_bts_trx *trx)
 {
 	int rc;
 
-	LOGPTRX(trx, DSUM, LOGL_INFO, "RSL link up\n");
+	LOGPTRX(trx, DRSL, LOGL_INFO, "RSL link up\n");
 
 	osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_RSL_UP, NULL);
 	osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_RSL_UP, NULL);
 
-	if ((rc = rsl_tx_rf_res(trx)) < 0)
-		oml_tx_failure_event_rep(&trx->bb_transc.mo, NM_SEVER_MAJOR, OSMO_EVT_MAJ_RSL_FAIL,
-					 "Failed to establish RSL link (%d)", rc);
-
+	if (trx->mo.nm_state.operational == NM_OPSTATE_ENABLED ||
+	    trx->bb_transc.mo.nm_state.operational == NM_OPSTATE_ENABLED) {
+		rc = rsl_tx_rf_res(trx);
+		if (rc < 0)
+			oml_tx_failure_event_rep(&trx->bb_transc.mo, NM_SEVER_MAJOR, OSMO_EVT_MAJ_RSL_FAIL,
+						 "Failed to establish RSL link (%d)", rc);
+	}
 	if (trx == trx->bts->c0)
 		load_timer_start(trx->bts);
 
