@@ -12,13 +12,15 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  */
 
+#include "btsconfig.h"	/* for PACKAGE_VERSION */
+
 #include <osmocom/core/logging.h>
 
 #include <osmocom/trau/osmo_ortp.h>
@@ -31,6 +33,7 @@
 #include <osmo-bts/handover.h>
 #include <osmo-bts/l1sap.h>
 #include <osmo-bts/bts_model.h>
+#include <osmo-bts/asci.h>
 #include <errno.h>
 
 static const struct value_string lchan_s_names[] = {
@@ -53,42 +56,55 @@ const struct value_string lchan_ciph_state_names[] = {
 	{ 0, NULL }
 };
 
+const struct value_string lchan_csd_mode_descs[] = {
+	{ LCHAN_CSD_M_NT,		"non-transparent" },
+	{ LCHAN_CSD_M_T_1200_75,	"transparent @ 1200/75 bps" },
+	{ LCHAN_CSD_M_T_600,		"transparent @ 600 bps" },
+	{ LCHAN_CSD_M_T_1200,		"transparent @ 1200 bps" },
+	{ LCHAN_CSD_M_T_2400,		"transparent @ 2400 bps" },
+	{ LCHAN_CSD_M_T_4800,		"transparent @ 4800 bps" },
+	{ LCHAN_CSD_M_T_9600,		"transparent @ 9600 bps" },
+	{ LCHAN_CSD_M_T_14400,		"transparent @ 14400 bps" },
+	{ LCHAN_CSD_M_T_29000,		"transparent @ 29000 bps" },
+	{ LCHAN_CSD_M_T_32000,		"transparent @ 32000 bps" },
+	{ 0, NULL }
+};
+
 /* prepare the per-SAPI T200 arrays for a given lchan */
-static int t200_by_lchan(int *t200_ms_dcch, int *t200_ms_acch, struct gsm_lchan *lchan)
+static int t200_by_lchan(uint32_t *t200_fn_dcch, uint32_t *t200_fn_acch, struct gsm_lchan *lchan)
 {
 	struct gsm_bts *bts = lchan->ts->trx->bts;
 
-	/* we have to compensate for the "RTS advance" due to the asynchronous interface between
-	 * the BTS (LAPDm) and the PHY/L1 (OsmoTRX or DSP in case of osmo-bts-{sysmo,lc15,oc2g,octphy} */
-	int32_t fn_advance = bts_get_avg_fn_advance(bts);
-	int32_t fn_advance_us = fn_advance * 4615;
-	int fn_advance_ms = fn_advance_us / 1000;
-
-	t200_ms_acch[DL_SAPI0] = bts->t200_ms[T200_SACCH_SDCCH] + fn_advance_ms;
-	t200_ms_acch[DL_SAPI3] = bts->t200_ms[T200_SACCH_SDCCH] + fn_advance_ms;
-
-	if (lchan->rep_acch_cap.dl_facch_all && lchan_is_tch(lchan)) {
-		t200_ms_acch[DL_SAPI0] *= 2;
-		t200_ms_acch[DL_SAPI3] *= 2;
-	}
-
 	switch (lchan->type) {
 	case GSM_LCHAN_SDCCH:
-		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_SDCCH] + fn_advance_ms;
-		t200_ms_dcch[DL_SAPI3] = bts->t200_ms[T200_SDCCH_SAPI3] + fn_advance_ms;
+		t200_fn_dcch[DL_SAPI0] = bts->t200_fn[T200_SDCCH];
+		t200_fn_dcch[DL_SAPI3] = bts->t200_fn[T200_SDCCH_SAPI3];
+		t200_fn_acch[DL_SAPI0] = bts->t200_fn[T200_SACCH_SDCCH];
+		t200_fn_acch[DL_SAPI3] = bts->t200_fn[T200_SACCH_SDCCH];
 		break;
 	case GSM_LCHAN_TCH_F:
-		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_FACCH_F] + fn_advance_ms;
-		t200_ms_dcch[DL_SAPI3] = bts->t200_ms[T200_FACCH_F] + fn_advance_ms;
+		t200_fn_dcch[DL_SAPI0] = bts->t200_fn[T200_FACCH_F];
+		t200_fn_dcch[DL_SAPI3] = bts->t200_fn[T200_FACCH_F];
+		t200_fn_acch[DL_SAPI0] = bts->t200_fn[T200_SACCH_TCH_SAPI0];
+		t200_fn_acch[DL_SAPI3] = bts->t200_fn[T200_SACCH_TCH_SAPI3];
 		break;
 	case GSM_LCHAN_TCH_H:
-		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_FACCH_H] + fn_advance_ms;
-		t200_ms_dcch[DL_SAPI3] = bts->t200_ms[T200_FACCH_H] + fn_advance_ms;
+		t200_fn_dcch[DL_SAPI0] = bts->t200_fn[T200_FACCH_H];
+		t200_fn_dcch[DL_SAPI3] = bts->t200_fn[T200_FACCH_H];
+		t200_fn_acch[DL_SAPI0] = bts->t200_fn[T200_SACCH_TCH_SAPI0];
+		t200_fn_acch[DL_SAPI3] = bts->t200_fn[T200_SACCH_TCH_SAPI3];
 		break;
 	default:
 		/* Channels such as CCCH don't use lapdm DL, and hence no T200 is needed */
 		return -1;
 	}
+
+	/* Add time of two extra messages frames. */
+	if (lchan->rep_acch_cap.dl_facch_all && lchan_is_tch(lchan)) {
+		t200_fn_acch[DL_SAPI0] += 104 * 2;
+		t200_fn_acch[DL_SAPI3] += 104 * 2;
+	}
+
 	return 0;
 }
 
@@ -130,6 +146,7 @@ void gsm_lchan_init(struct gsm_lchan *lchan, struct gsm_bts_trx_ts *ts, unsigned
 
 	INIT_LLIST_HEAD(&lchan->sapi_cmds);
 	INIT_LLIST_HEAD(&lchan->dl_tch_queue);
+	lchan->dl_tch_queue_len = 0;
 }
 
 void gsm_lchan_name_update(struct gsm_lchan *lchan)
@@ -148,17 +165,17 @@ void gsm_lchan_name_update(struct gsm_lchan *lchan)
 int lchan_init_lapdm(struct gsm_lchan *lchan)
 {
 	struct lapdm_channel *lc = &lchan->lapdm_ch;
-	int t200_ms_dcch[_NR_DL_SAPI], t200_ms_acch[_NR_DL_SAPI];
+	uint32_t t200_fn_dcch[_NR_DL_SAPI], t200_fn_acch[_NR_DL_SAPI];
 
-	if (t200_by_lchan(t200_ms_dcch, t200_ms_acch, lchan) == 0) {
+	if (t200_by_lchan(t200_fn_dcch, t200_fn_acch, lchan) == 0) {
 		LOGPLCHAN(lchan, DLLAPD, LOGL_DEBUG,
-			  "Setting T200 D0=%u, D3=%u, S0=%u, S3=%u (all in ms)\n",
-			  t200_ms_dcch[DL_SAPI0], t200_ms_dcch[DL_SAPI3],
-			  t200_ms_acch[DL_SAPI0], t200_ms_acch[DL_SAPI3]);
-		lapdm_channel_init3(lc, LAPDM_MODE_BTS, t200_ms_dcch, t200_ms_acch, lchan->type,
-				    gsm_lchan_name(lchan));
-		lapdm_channel_set_flags(lc, LAPDM_ENT_F_POLLING_ONLY);
+			  "Setting T200 D0=%u, D3=%u, S0=%u, S3=%u (all in frames)\n",
+			  t200_fn_dcch[DL_SAPI0], t200_fn_dcch[DL_SAPI3],
+			  t200_fn_acch[DL_SAPI0], t200_fn_acch[DL_SAPI3]);
+		lapdm_channel_init3(lc, LAPDM_MODE_BTS, NULL, NULL, lchan->type, gsm_lchan_name(lchan));
+		lapdm_channel_set_flags(lc, LAPDM_ENT_F_POLLING_ONLY | LAPDM_ENT_F_RTS);
 		lapdm_channel_set_l1(lc, NULL, lchan);
+		lapdm_channel_set_t200_fn(lc, t200_fn_dcch, t200_fn_acch);
 	}
 	/* We still need to set Rx callback to receive RACH requests: */
 	lapdm_channel_set_l3(lc, lapdm_rll_tx_cb, lchan);
@@ -200,10 +217,12 @@ void gsm_lchan_release(struct gsm_lchan *lchan, enum lchan_rel_act_kind rel_kind
 		rsl_tx_ipac_dlcx_ind(lchan, RSL_ERR_NORMAL_UNSPEC);
 		osmo_rtp_socket_log_stats(lchan->abis_ip.rtp_socket, DRTP, LOGL_INFO,
 			"Closing RTP socket on Channel Release ");
-		osmo_rtp_socket_free(lchan->abis_ip.rtp_socket);
-		lchan->abis_ip.rtp_socket = NULL;
-		msgb_queue_flush(&lchan->dl_tch_queue);
+		lchan_rtp_socket_free(lchan);
+	} else if (lchan->abis_ip.osmux.use) {
+		lchan_osmux_release(lchan);
 	}
+	/* reset all Abis related config: */
+	memset(&lchan->abis_ip, 0, sizeof(lchan->abis_ip));
 
 	/* FIXME: right now we allow creating the rtp_socket even if chan is not
 	 * activated... Once we check for that, we can move this check at the
@@ -211,8 +230,11 @@ void gsm_lchan_release(struct gsm_lchan *lchan, enum lchan_rel_act_kind rel_kind
 	if (lchan->state == LCHAN_S_NONE)
 		return;
 
-	/* release handover state */
+	/* release handover, listener and talker states */
 	handover_reset(lchan);
+	vgcs_talker_reset(lchan, false);
+	vgcs_listener_reset(lchan);
+	vgcs_uplink_free_reset(lchan);
 
 	lchan->rel_act_kind = rel_kind;
 
@@ -317,7 +339,7 @@ void lchan_set_state(struct gsm_lchan *lchan, enum gsm_lchan_state state)
 	}
 }
 
-/* See Table 10.5.25 of GSM04.08 */
+/* See 3GPP TS 44.018 Table 10.5.2.5.1 "Channel Description information element" */
 static uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
 			  uint8_t ts_nr, uint8_t lchan_nr)
 {
@@ -362,13 +384,13 @@ static uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
 		cbits = ABIS_RSL_CHAN_NR_CBITS_BCCH;
 		break;
 	case GSM_PCHAN_NONE:
-		LOGP(DRSL, LOGL_ERROR, "Physical channel %s not expected!\n",
-		     gsm_pchan_name(pchan));
-		cbits = 0x00;
+		LOGP(DRSL, LOGL_ERROR, "ts=%u,ss=%u Physical channel %s not expected!\n",
+		     ts_nr, lchan_nr, gsm_pchan_name(pchan));
+		cbits = 0x00; /* Reserved */
 		break;
 	default:
-		LOGP(DRSL, LOGL_ERROR, "Physical channel %s (0x%02x) not expected!\n",
-		     gsm_pchan_name(pchan), (int)pchan);
+		LOGP(DRSL, LOGL_ERROR, "ts=%u,ss=%u Physical channel %s (0x%02x) not expected!\n",
+		     ts_nr, lchan_nr, gsm_pchan_name(pchan), (int)pchan);
 		OSMO_ASSERT(0);
 		break;
 	}
@@ -510,3 +532,148 @@ int lchan2ecu_codec(const struct gsm_lchan *lchan)
 		return -1;
 	}
 }
+
+static int bind_rtp(struct gsm_bts *bts, struct osmo_rtp_socket *rs, const char *ip)
+{
+	int rc;
+	unsigned int i;
+	unsigned int tries;
+
+	tries = (bts->rtp_port_range_end - bts->rtp_port_range_start) / 2;
+	for (i = 0; i < tries; i++) {
+
+		if (bts->rtp_port_range_next >= bts->rtp_port_range_end)
+			bts->rtp_port_range_next = bts->rtp_port_range_start;
+
+		rc = osmo_rtp_socket_bind(rs, ip, bts->rtp_port_range_next);
+
+		bts->rtp_port_range_next += 2;
+
+		if (rc != 0)
+			continue;
+
+		if (bts->rtp_ip_dscp != -1) {
+			if (osmo_rtp_socket_set_dscp(rs, bts->rtp_ip_dscp))
+				LOGP(DRSL, LOGL_ERROR, "failed to set DSCP=%d: %s\n",
+					bts->rtp_ip_dscp, strerror(errno));
+		}
+		if (bts->rtp_priority != -1) {
+			if (osmo_rtp_socket_set_priority(rs, bts->rtp_priority))
+				LOGP(DRSL, LOGL_ERROR, "failed to set socket priority %d: %s\n",
+					bts->rtp_priority, strerror(errno));
+		}
+		return 0;
+	}
+
+	return -1;
+}
+
+int lchan_rtp_socket_create(struct gsm_lchan *lchan, const char *bind_ip)
+{
+	struct gsm_bts *bts = lchan->ts->trx->bts;
+	char cname[256+4];
+	int rc;
+
+	if (lchan->abis_ip.rtp_socket) {
+		LOGPLCHAN(lchan, DRSL, LOGL_ERROR, "Rx RSL IPAC CRCX, "
+			  "but we already have socket!\n");
+		return -EALREADY;
+	}
+
+	/* FIXME: select default value depending on speech_mode */
+	//if (!payload_type)
+	lchan->tch.last_fn = LCHAN_FN_DUMMY;
+	lchan->abis_ip.rtp_socket = osmo_rtp_socket_create(lchan->ts->trx,
+							OSMO_RTP_F_POLL);
+
+	if (!lchan->abis_ip.rtp_socket) {
+		LOGPLCHAN(lchan, DRTP, LOGL_ERROR, "IPAC Failed to create RTP/RTCP sockets\n");
+		oml_tx_failure_event_rep(&lchan->ts->trx->mo,
+					 NM_SEVER_MINOR, OSMO_EVT_CRIT_RTP_TOUT,
+					 "%s IPAC Failed to create RTP/RTCP sockets",
+					 gsm_lchan_name(lchan));
+		return -ENOTCONN;
+	}
+
+	rc = osmo_rtp_socket_set_param(lchan->abis_ip.rtp_socket,
+				       bts->rtp_jitter_adaptive ?
+				       OSMO_RTP_P_JIT_ADAP :
+				       OSMO_RTP_P_JITBUF,
+				       bts->rtp_jitter_buf_ms);
+	if (rc < 0)
+		LOGPLCHAN(lchan, DRTP, LOGL_ERROR,
+			  "IPAC Failed to set RTP socket parameters: %s\n", strerror(-rc));
+	else
+		LOGPLCHAN(lchan, DRTP, LOGL_INFO, "IPAC set RTP socket parameters: %d\n", rc);
+
+	lchan->abis_ip.rtp_socket->priv = lchan;
+	lchan->abis_ip.rtp_socket->rx_cb = &l1sap_rtp_rx_cb;
+
+	rc = bind_rtp(bts, lchan->abis_ip.rtp_socket, bind_ip);
+	if (rc < 0) {
+		LOGPLCHAN(lchan, DRTP, LOGL_ERROR, "IPAC Failed to bind RTP/RTCP sockets\n");
+		oml_tx_failure_event_rep(&lchan->ts->trx->mo,
+					 NM_SEVER_MINOR, OSMO_EVT_CRIT_RTP_TOUT,
+					 "%s IPAC Failed to bind RTP/RTCP sockets",
+					 gsm_lchan_name(lchan));
+		lchan_rtp_socket_free(lchan);
+		return -EBADFD;
+	}
+
+	/* Ensure RTCP SDES contains some useful information */
+	snprintf(cname, sizeof(cname), "bts@%s", bind_ip);
+	osmo_rtp_set_source_desc(lchan->abis_ip.rtp_socket, cname,
+				 gsm_lchan_name(lchan), NULL, NULL,
+				 gsm_trx_unit_id(lchan->ts->trx),
+				 "OsmoBTS-" PACKAGE_VERSION, NULL);
+	/* FIXME: multiplex connection, BSC proxy */
+	return 0;
+}
+
+
+int lchan_rtp_socket_connect(struct gsm_lchan *lchan, const struct in_addr *ia, uint16_t connect_port)
+{
+	int bound_port = 0;
+	int rc;
+
+	rc = osmo_rtp_socket_connect(lchan->abis_ip.rtp_socket,
+				     inet_ntoa(*ia), connect_port);
+	if (rc < 0) {
+		LOGPLCHAN(lchan, DRTP, LOGL_ERROR, "Failed to connect RTP/RTCP sockets\n");
+		return -ECONNREFUSED;
+	}
+	/* save IP address and port number */
+	lchan->abis_ip.connect_ip = ntohl(ia->s_addr);
+	lchan->abis_ip.connect_port = connect_port;
+
+	rc = osmo_rtp_get_bound_ip_port(lchan->abis_ip.rtp_socket,
+					&lchan->abis_ip.bound_ip,
+					&bound_port);
+	if (rc < 0)
+		LOGPLCHAN(lchan, DRTP, LOGL_ERROR, "IPAC cannot obtain locally bound IP/port: %d\n", rc);
+	lchan->abis_ip.bound_port = bound_port;
+	return 0;
+}
+
+void lchan_rtp_socket_free(struct gsm_lchan *lchan)
+{
+	osmo_rtp_socket_free(lchan->abis_ip.rtp_socket);
+	lchan->abis_ip.rtp_socket = NULL;
+	msgb_queue_free(&lchan->dl_tch_queue);
+	lchan->dl_tch_queue_len = 0;
+}
+
+/*! limit number of queue entries to %u; drops any surplus messages */
+void lchan_dl_tch_queue_enqueue(struct gsm_lchan *lchan, struct msgb *msg, unsigned int limit)
+{
+	if (lchan->dl_tch_queue_len > limit) {
+		unsigned int excess = lchan->dl_tch_queue_len - limit;
+		LOGPLCHAN(lchan, DL1P, LOGL_NOTICE, "freeing %d queued frames\n", excess);
+		rate_ctr_add2(lchan->ts->trx->bts->ctrs, BTS_CTR_RTP_RX_DROP_OVERFLOW, excess);
+	}
+	while (lchan->dl_tch_queue_len > limit) {
+		struct msgb *tmp = msgb_dequeue_count(&lchan->dl_tch_queue, &lchan->dl_tch_queue_len);
+		msgb_free(tmp);
+	}
+	msgb_enqueue_count(&lchan->dl_tch_queue, msg, &lchan->dl_tch_queue_len);
+}
