@@ -2,19 +2,40 @@
 #include <stdint.h>
 #include <errno.h>
 
-#include <osmocom/gsm/gsm_utils.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/endian.h>
+
+#include <osmocom/gsm/gsm_utils.h>
+#include <osmocom/gsm/protocol/gsm_44_004.h>
 
 #include <osmo-bts/gsm_data.h>
 #include <osmo-bts/logging.h>
 #include <osmo-bts/measurement.h>
 #include <osmo-bts/scheduler.h>
 #include <osmo-bts/rsl.h>
+#include <osmo-bts/power_control.h>
+#include <osmo-bts/ta_control.h>
 
-/* Tables as per TS 45.008 Section 8.3 */
-static const uint8_t ts45008_83_tch_f[] = { 52, 53, 54, 55, 56, 57, 58, 59 };
-static const uint8_t ts45008_83_tch_hs0[] = { 0, 2, 4, 6, 52, 54, 56, 58 };
-static const uint8_t ts45008_83_tch_hs1[] = { 14, 16, 18, 20, 66, 68, 70, 72 };
+/* Active TDMA frame subset for TCH/H in DTX mode (see 3GPP TS 45.008 Section 8.3).
+ * This mapping is used to determine if a L2 block starting at the given TDMA FN
+ * belongs to the SUB set and thus shall always be transmitted in DTX mode. */
+static const uint8_t ts45008_dtx_tchh_speech_fn_map[104] = {
+	/* TCH/H(0): 0, 2, 4, 6, 52, 54, 56, 58 */
+	[0]  = 1, /* block { 0,  2,  4,  6} */
+	[52] = 1, /* block {52, 54, 56, 58} */
+	/* TCH/H(1): 14, 16, 18, 20, 66, 68, 70, 72 */
+	[14] = 1, /* block {14, 16, 18, 20} */
+	[66] = 1, /* block {66, 68, 70, 72} */
+};
+
+static const uint8_t ts45008_dtx_tchh_data_fn_map[104] = {
+	/* UL TCH/H(0): 52, 54, 56, 58, 60, 62, 65, 67, 69, 71 */
+	[52] = 1, /* block {52, 54, 56, 58, 60, 62} */
+	[60] = 1, /* block {60, 62, 65, 67, 69, 71} */
+	/* UL TCH/H(1): 70, 72, 74, 76, 79, 81, 83, 85, 87, 89 */
+	[70] = 1, /* block {70, 72, 74, 76, 79, 81} */
+	[79] = 1, /* block {79, 81, 83, 85, 87, 89} */
+};
 
 /* In cases where we less measurements than we expect we must assume that we
  * just did not receive the block because it was lost due to bad channel
@@ -23,88 +44,74 @@ static const uint8_t ts45008_83_tch_hs1[] = { 14, 16, 18, 20, 66, 68, 70, 72 };
  * the missing measurements */
 #define MEASUREMENT_DUMMY_BER 10000 /* 100% BER */
 #define MEASUREMENT_DUMMY_IRSSI 109 /* noise floor in -dBm */
-static const struct bts_ul_meas measurement_dummy = (struct bts_ul_meas) {
+static const struct bts_ul_meas measurement_dummy = {
 	.ber10k = MEASUREMENT_DUMMY_BER,
 	.ta_offs_256bits = 0,
-	.c_i = 0,
+	.ci_cb = 0,
 	.is_sub = 0,
 	.inv_rssi = MEASUREMENT_DUMMY_IRSSI
 };
 
-/* find out if an array contains a given key as element */
-#define ARRAY_CONTAINS(arr, val) array_contains(arr, ARRAY_SIZE(arr), val)
-static bool array_contains(const uint8_t *arr, unsigned int len, uint8_t val) {
-	int i;
-	for (i = 0; i < len; i++) {
-		if (arr[i] == val)
-			return true;
-	}
-	return false;
-}
-
 /* Decide if a given frame number is part of the "-SUB" measurements (true) or not (false)
  * (this function is only used internally, it is public to call it from unit-tests) */
-bool ts45008_83_is_sub(struct gsm_lchan *lchan, uint32_t fn, bool is_amr_sid_update)
+bool ts45008_83_is_sub(struct gsm_lchan *lchan, uint32_t fn)
 {
 	uint32_t fn104 = fn % 104;
 
 	/* See TS 45.008 Sections 8.3 and 8.4 for a detailed descriptions of the rules
-	 * implemented here. We only implement the logic for Voice, not CSD */
+	 * implemented here. We implement the logic for both speech and data (CSD). */
+
+	/* AMR is special, SID frames may be scheduled dynamically at any time */
+	if (lchan->tch_mode == GSM48_CMODE_SPEECH_AMR)
+		return false;
 
 	switch (lchan->type) {
 	case GSM_LCHAN_TCH_F:
 		switch (lchan->tch_mode) {
-		case GSM48_CMODE_SIGN:
+		case GSM48_CMODE_SIGN: /* TCH/F sign: DTX *is* permitted */
 		case GSM48_CMODE_SPEECH_V1:
+		case GSM48_CMODE_SPEECH_V1_VAMOS:
 		case GSM48_CMODE_SPEECH_EFR:
-			if (trx_sched_is_sacch_fn(lchan->ts, fn, true))
-				return true;
-			if (ARRAY_CONTAINS(ts45008_83_tch_f, fn104))
-				return true;
-			break;
-		case GSM48_CMODE_SPEECH_AMR:
-			if (trx_sched_is_sacch_fn(lchan->ts, fn, true))
-				return true;
-			if (is_amr_sid_update)
+		case GSM48_CMODE_SPEECH_V2_VAMOS:
+			/* Active TDMA frame subset for TCH/F: 52, 53, 54, 55, 56, 57, 58, 59.
+			 * There is only one *complete* block in this subset starting at FN=52.
+			 * Incomplete blocks {... 52, 53, 54, 55} and {56, 57, 58, 59 ...}
+			 * contain only 50% of the useful bits (partial SID) and thus ~50% BER. */
+			if (fn104 == 52)
 				return true;
 			break;
+		case GSM48_CMODE_DATA_12k0: /* TCH/F9.6 */
+		case GSM48_CMODE_DATA_6k0: /* TCH/F4.8 */
+			/* FIXME: The RXQUAL_SUB (not RXLEV!) report shall include measurements on
+			 * the TDMA frames given in the table of subclause 8.3 only if L2 fill frames
+			 * have been received as FACCH/F frames at the corresponding frame positions. */
 		default:
-			LOGPFN(DMEAS, LOGL_ERROR, fn, "%s: Unsupported lchan->tch_mode %u\n",
-				gsm_lchan_name(lchan), lchan->tch_mode);
+			if (lchan->rsl_cmode == RSL_CMOD_SPD_DATA)
+				return fn104 == 52;
+			LOGPLCFN(lchan, fn, DMEAS, LOGL_ERROR, "Unsupported lchan->tch_mode %u\n", lchan->tch_mode);
 			break;
 		}
 		break;
 	case GSM_LCHAN_TCH_H:
 		switch (lchan->tch_mode) {
 		case GSM48_CMODE_SPEECH_V1:
-			if (trx_sched_is_sacch_fn(lchan->ts, fn, true))
-				return true;
-			switch (lchan->nr) {
-			case 0:
-				if (ARRAY_CONTAINS(ts45008_83_tch_hs0, fn104))
-					return true;
-				break;
-			case 1:
-				if (ARRAY_CONTAINS(ts45008_83_tch_hs1, fn104))
-					return true;
-				break;
-			default:
-				OSMO_ASSERT(0);
-			}
-			break;
-		case GSM48_CMODE_SPEECH_AMR:
-			if (trx_sched_is_sacch_fn(lchan->ts, fn, true))
-				return true;
-			if (is_amr_sid_update)
+		case GSM48_CMODE_SPEECH_V1_VAMOS:
+			if (ts45008_dtx_tchh_speech_fn_map[fn104])
 				return true;
 			break;
 		case GSM48_CMODE_SIGN:
 			/* No DTX allowed; SUB=FULL, therefore measurements at all frame numbers are
 			 * SUB */
 			return true;
+		case GSM48_CMODE_DATA_6k0: /* TCH/H4.8 */
+		case GSM48_CMODE_DATA_3k6: /* TCH/H2.4 */
+			/* FIXME: The RXQUAL_SUB (not RXLEV!) report shall include measurements on
+			 * the TDMA frames given in the table of subclause 8.3 only if L2 fill frames
+			 * have been received as FACCH/H frames at the corresponding frame positions. */
 		default:
-			LOGPFN(DMEAS, LOGL_ERROR, fn, "%s: Unsupported lchan->tch_mode %u\n",
-				gsm_lchan_name(lchan), lchan->tch_mode);
+			if (lchan->rsl_cmode == RSL_CMOD_SPD_DATA)
+				return ts45008_dtx_tchh_data_fn_map[fn104] == 1;
+			LOGPLCFN(lchan, fn, DMEAS, LOGL_ERROR, "Unsupported lchan->tch_mode %u\n", lchan->tch_mode);
 			break;
 		}
 		break;
@@ -250,11 +257,6 @@ int is_meas_complete(struct gsm_lchan *lchan, uint32_t fn)
 	int rc = 0;
 	enum gsm_phys_chan_config pchan = ts_pchan(lchan->ts);
 
-	if (lchan->ts->nr >= 8)
-		return -EINVAL;
-	if (pchan >= _GSM_PCHAN_MAX)
-		return -EINVAL;
-
 	switch (pchan) {
 	case GSM_PCHAN_TCH_F:
 		fn_mod = translate_tch_meas_rep_fn104(fn % 104);
@@ -288,9 +290,8 @@ int is_meas_complete(struct gsm_lchan *lchan, uint32_t fn)
 	}
 
 	if (rc == 1) {
-		DEBUGP(DMEAS,
-		       "%s meas period end fn:%u, fn_mod:%i, status:%d, pchan:%s\n",
-		       gsm_lchan_name(lchan), fn, fn_mod, rc, gsm_pchan_name(pchan));
+		LOGPLCFN(lchan, fn, DMEAS, LOGL_DEBUG, "meas period end fn_mod:%d, status:%d, pchan:%s\n", fn_mod,
+			 rc, gsm_pchan_name(pchan));
 	}
 
 	return rc;
@@ -305,50 +306,51 @@ static uint8_t modulus_by_lchan(struct gsm_lchan *lchan)
 	case GSM_PCHAN_TCH_F:
 	case GSM_PCHAN_TCH_H:
 		return 104;
-		break;
 	case GSM_PCHAN_SDCCH8_SACCH8C:
 	case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
 	case GSM_PCHAN_CCCH_SDCCH4:
 	case GSM_PCHAN_CCCH_SDCCH4_CBCH:
 		return 102;
-		break;
 	default:
 		/* Invalid */
 		return 1;
-		break;
 	}
 }
 
 /* receive a L1 uplink measurement from L1 (this function is only used
  * internally, it is public to call it from unit-tests)  */
-int lchan_new_ul_meas(struct gsm_lchan *lchan, struct bts_ul_meas *ulm, uint32_t fn)
+int lchan_new_ul_meas(struct gsm_lchan *lchan,
+		      const struct bts_ul_meas *ulm,
+		      uint32_t fn)
 {
 	uint32_t fn_mod = fn % modulus_by_lchan(lchan);
+	struct bts_ul_meas *dest;
 
 	if (lchan->state != LCHAN_S_ACTIVE) {
-		LOGPFN(DMEAS, LOGL_NOTICE, fn,
-		       "%s measurement during state: %s, num_ul_meas=%d, fn_mod=%u\n",
-		       gsm_lchan_name(lchan), gsm_lchans_name(lchan->state),
-		       lchan->meas.num_ul_meas, fn_mod);
+		LOGPLCFN(lchan, fn, DMEAS, LOGL_NOTICE,
+			 "measurement during state: %s, num_ul_meas=%d, fn_mod=%u\n",
+			 gsm_lchans_name(lchan->state), lchan->meas.num_ul_meas, fn_mod);
 	}
 
 	if (lchan->meas.num_ul_meas >= ARRAY_SIZE(lchan->meas.uplink)) {
-		LOGPFN(DMEAS, LOGL_NOTICE, fn,
-		       "%s no space for uplink measurement, num_ul_meas=%d, fn_mod=%u\n",
-		       gsm_lchan_name(lchan), lchan->meas.num_ul_meas, fn_mod);
+		LOGPLCFN(lchan, fn, DMEAS, LOGL_NOTICE,
+			 "no space for uplink measurement, num_ul_meas=%d, fn_mod=%u\n", lchan->meas.num_ul_meas,
+			 fn_mod);
 		return -ENOSPC;
 	}
 
+	dest = &lchan->meas.uplink[lchan->meas.num_ul_meas++];
+	memcpy(dest, ulm, sizeof(*ulm));
+
 	/* We expect the lower layers to mark AMR SID_UPDATE frames already as such.
 	 * In this function, we only deal with the common logic as per the TS 45.008 tables */
 	if (!ulm->is_sub)
-		ulm->is_sub = ts45008_83_is_sub(lchan, fn, false);
+		dest->is_sub = ts45008_83_is_sub(lchan, fn);
 
-	DEBUGPFN(DMEAS, fn, "%s adding measurement (is_sub=%u), num_ul_meas=%d, fn_mod=%u\n",
-		 gsm_lchan_name(lchan), ulm->is_sub, lchan->meas.num_ul_meas, fn_mod);
-
-	memcpy(&lchan->meas.uplink[lchan->meas.num_ul_meas++], ulm,
-		sizeof(*ulm));
+	LOGPLCFN(lchan, fn, DMEAS, LOGL_DEBUG,
+		 "adding a %s measurement (ber10k=%u, ta_offs=%d, ci_cB=%d, rssi=-%u), num_ul_meas=%d, fn_mod=%u\n",
+		 dest->is_sub ? "SUB" : "FULL", ulm->ber10k, ulm->ta_offs_256bits, ulm->ci_cb, ulm->inv_rssi,
+		 lchan->meas.num_ul_meas, fn_mod);
 
 	lchan->meas.last_fn = fn;
 
@@ -398,11 +400,16 @@ static unsigned int lchan_meas_num_expected(const struct gsm_lchan *lchan)
 
 	switch (pchan) {
 	case GSM_PCHAN_TCH_F:
-		/* 24 for TCH + 1 for SACCH */
+		/* 24 blocks for TCH + 1 for SACCH */
 		return 25;
 	case GSM_PCHAN_TCH_H:
-		/* 24 half-blocks for TCH + 1 for SACCH */
-		return 25;
+		if (lchan->tch_mode == GSM48_CMODE_SIGN) {
+			/* 12 blocks for TCH + 1 for SACCH */
+			return 13;
+		} else {
+			/* 24 blocks for TCH + 1 for SACCH */
+			return 25;
+		}
 	case GSM_PCHAN_SDCCH8_SACCH8C:
 	case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
 		/* 2 for SDCCH + 1 for SACCH */
@@ -417,23 +424,51 @@ static unsigned int lchan_meas_num_expected(const struct gsm_lchan *lchan)
 }
 
 /* In DTX a subset of blocks must always be transmitted
- * See also: GSM 05.08, chapter 8.3 Aspects of discontinuous transmission (DTX) */
-static unsigned int lchan_meas_sub_num_expected(const struct gsm_lchan *lchan)
+ * See also: GSM 05.08, chapter 8.3 Aspects of discontinuous transmission (DTX)
+ * Return value N: (N < 0) -- at least N SUB frames expected;
+ *                 (N > 0) -- exactly N SUB frames expected;
+ *                 (N == 0) - unknown channel type/mode? */
+static int lchan_meas_sub_num_expected(const struct gsm_lchan *lchan)
 {
 	enum gsm_phys_chan_config pchan = ts_pchan(lchan->ts);
 
-	/* AMR is using a more elaborated model with a dymanic amount of
-	 * DTX blocks so this function is not applicable to determine the
-	 * amount of expected SUB Measurements when AMR is used */
-	OSMO_ASSERT(lchan->tch_mode != GSM48_CMODE_SPEECH_AMR)
-
 	switch (pchan) {
 	case GSM_PCHAN_TCH_F:
-		/* 1 block SDCCH, 2 blocks TCH */
-		return 3;
+		if (lchan->rsl_cmode == RSL_CMOD_SPD_DATA)
+			return 1 + 1; /* 1 x SACCH + 1 x FACCH */
+		/* else: signalling or speech */
+		switch (lchan->tch_mode) {
+		case GSM48_CMODE_SIGN: /* TCH/F sign: DTX *is* permitted */
+		case GSM48_CMODE_SPEECH_V1: /* TCH/FS */
+		case GSM48_CMODE_SPEECH_V1_VAMOS:
+		case GSM48_CMODE_SPEECH_EFR: /* TCH/EFS */
+		case GSM48_CMODE_SPEECH_V2_VAMOS:
+			return 1 + 1; /* 1 x SACCH + 1 x TCH */
+		case GSM48_CMODE_SPEECH_AMR: /* TCH/AFS */
+		case GSM48_CMODE_SPEECH_V3_VAMOS:
+		case GSM48_CMODE_SPEECH_V4: /* O-TCH/WFS */
+		case GSM48_CMODE_SPEECH_V5: /* TCH/WFS */
+		case GSM48_CMODE_SPEECH_V5_VAMOS:
+		default:
+			return -1; /* at least 1 x SACCH + M x TCH (variable) */
+		}
 	case GSM_PCHAN_TCH_H:
-		/* 1 block SDCCH, 4 half-blocks TCH */
-		return 5;
+		if (lchan->rsl_cmode == RSL_CMOD_SPD_DATA)
+			return 1 + 2; /* 1 x SACCH + 2 x FACCH */
+		/* else: signalling or speech */
+		switch (lchan->tch_mode) {
+		case GSM48_CMODE_SIGN: /* TCH/H sign: DTX *is not* permitted */
+			return 1 + 12; /* 1 x SACCH + 12 x TCH */
+		case GSM48_CMODE_SPEECH_V1:
+		case GSM48_CMODE_SPEECH_V1_VAMOS:
+			return 1 + 2; /* 1 x SACCH + 2 x TCH */
+		case GSM48_CMODE_SPEECH_AMR: /* TCH/AHS */
+		case GSM48_CMODE_SPEECH_V3_VAMOS:
+		case GSM48_CMODE_SPEECH_V4: /* O-TCH/WHS */
+		case GSM48_CMODE_SPEECH_V6: /* O-TCH/AHS */
+		default:
+			return -1; /* at least 1 x SACCH + M x TCH (variable) */
+		}
 	case GSM_PCHAN_SDCCH8_SACCH8C:
 	case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
 		/* no DTX here, all blocks must be present! */
@@ -471,7 +506,7 @@ static void lchan_meas_compute_extended(struct gsm_lchan *lchan)
 
 	/* each measurement is an int32_t, so the squared difference value must fit in 32bits */
 	/* the sum of the squared values (each up to 32bit) can very easily exceed 32 bits */
-	u_int64_t sq_diff_sum = 0;
+	uint64_t sq_diff_sum = 0;
 
 	/* In case we do not have any measurement values collected there is no
 	 * computation possible. We just skip the whole computation here, the
@@ -542,13 +577,15 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 	struct gsm_meas_rep_unidir *mru;
 	uint32_t ber_full_sum = 0;
 	uint32_t irssi_full_sum = 0;
+	int32_t ci_full_sum = 0;
 	uint32_t ber_sub_sum = 0;
 	uint32_t irssi_sub_sum = 0;
+	int32_t ci_sub_sum = 0;
 	int32_t ta256b_sum = 0;
 	unsigned int num_meas_sub = 0;
 	unsigned int num_meas_sub_actual = 0;
 	unsigned int num_meas_sub_subst = 0;
-	unsigned int num_meas_sub_expect;
+	int num_meas_sub_expect;
 	unsigned int num_ul_meas;
 	unsigned int num_ul_meas_actual = 0;
 	unsigned int num_ul_meas_subst = 0;
@@ -560,33 +597,25 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 	if (!is_meas_complete(lchan, fn))
 		return 0;
 
-	LOGP(DMEAS, LOGL_DEBUG, "%s Calculating measurement results for physical channel:%s\n",
-	     gsm_lchan_name(lchan), gsm_pchan_name(ts_pchan(lchan->ts)));
+	LOGPLCHAN(lchan, DMEAS, LOGL_DEBUG, "Calculating measurement results "
+		  "for physical channel: %s\n", gsm_pchan_name(ts_pchan(lchan->ts)));
 
 	/* Note: Some phys will send no measurement indication at all
 	 * when a block is lost. Also in DTX mode blocks are left out
 	 * intentionally to save energy. It is not necessarly an error
 	 * when we get less measurements as we expect. */
 	num_ul_meas_expect = lchan_meas_num_expected(lchan);
-
-	if (lchan->tch_mode != GSM48_CMODE_SPEECH_AMR)
-		num_meas_sub_expect = lchan_meas_sub_num_expected(lchan);
-	else {
-		/* FIXME: the amount of SUB Measurements is a dynamic parameter
-		 * in AMR and can not be determined by using a lookup table.
-		 * See also: OS#2978 */
-		num_meas_sub_expect = 0;
-	}
+	num_meas_sub_expect = lchan_meas_sub_num_expected(lchan);
 
 	if (lchan->meas.num_ul_meas > num_ul_meas_expect)
 		num_ul_meas_excess = lchan->meas.num_ul_meas - num_ul_meas_expect;
 	num_ul_meas = num_ul_meas_expect;
 
-	LOGP(DMEAS, LOGL_DEBUG, "%s received %u UL measurements, expected %u\n", gsm_lchan_name(lchan),
-	     lchan->meas.num_ul_meas, num_ul_meas_expect);
+	LOGPLCHAN(lchan, DMEAS, LOGL_DEBUG, "Received %u UL measurements, expected %u\n",
+		  lchan->meas.num_ul_meas, num_ul_meas_expect);
 	if (num_ul_meas_excess)
-		LOGP(DMEAS, LOGL_DEBUG, "%s received %u excess UL measurements\n", gsm_lchan_name(lchan),
-		     num_ul_meas_excess);
+		LOGPLCHAN(lchan, DMEAS, LOGL_DEBUG, "Received %u excess UL measurements\n",
+			  num_ul_meas_excess);
 
 	/* Measurement computation step 1: add up */
 	for (i = 0; i < num_ul_meas; i++) {
@@ -610,18 +639,24 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 			m = &lchan->meas.uplink[i + num_ul_meas_excess];
 			if (m->is_sub) {
 				irssi_sub_sum += m->inv_rssi;
+				ci_sub_sum += m->ci_cb;
 				num_meas_sub_actual++;
 				is_sub = true;
 			}
 			irssi_full_sum += m->inv_rssi;
 			ta256b_sum += m->ta_offs_256bits;
+			ci_full_sum += m->ci_cb;
 
 			num_ul_meas_actual++;
 		} else {
 			m = &measurement_dummy;
-			if (num_ul_meas_expect - i <= num_meas_sub_expect - num_meas_sub) {
-				num_meas_sub_subst++;
-				is_sub = true;
+
+			/* only if we know the exact number of SUB measurements */
+			if (num_meas_sub_expect >= 0) {
+				if (num_meas_sub < num_meas_sub_expect) {
+					num_meas_sub_subst++;
+					is_sub = true;
+				}
 			}
 
 			num_ul_meas_subst++;
@@ -634,50 +669,71 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 		}
 	}
 
-	LOGP(DMEAS, LOGL_DEBUG, "%s received UL measurements contain %u SUB measurements, expected %u\n",
-	     gsm_lchan_name(lchan), num_meas_sub_actual, num_meas_sub_expect);
-	LOGP(DMEAS, LOGL_DEBUG, "%s replaced %u measurements with dummy values, from which %u were SUB measurements\n",
-	     gsm_lchan_name(lchan), num_ul_meas_subst, num_meas_sub_subst);
+	LOGPLCHAN(lchan, DMEAS, LOGL_DEBUG, "Replaced %u measurements with dummy values, "
+		  "from which %u were SUB measurements\n", num_ul_meas_subst, num_meas_sub_subst);
 
-	if (num_meas_sub != num_meas_sub_expect) {
-		LOGP(DMEAS, LOGL_ERROR, "%s Incorrect number of SUB measurements detected! (%u vs exp %u)\n",
-		     gsm_lchan_name(lchan), num_meas_sub, num_meas_sub_expect);
-		/* Normally the logic above should make sure that there is
-		 * always the exact amount of SUB measurements taken into
-		 * account. If not then the logic that decides tags the received
-		 * measurements as is_sub works incorrectly. Since the logic
-		 * above only adds missing measurements during the calculation
-		 * it can not remove excess SUB measurements or add missing SUB
-		 * measurements when there is no more room in the interval. */
+	/* Normally the logic above should make sure that there is
+	 * always the exact amount of SUB measurements taken into
+	 * account. If not then the logic that decides tags the received
+	 * measurements as is_sub works incorrectly. Since the logic
+	 * above only adds missing measurements during the calculation
+	 * it can not remove excess SUB measurements or add missing SUB
+	 * measurements when there is no more room in the interval. */
+	if (num_meas_sub_expect < 0) {
+		num_meas_sub_expect = -num_meas_sub_expect;
+		LOGPLCHAN(lchan, DMEAS, LOGL_DEBUG,
+			  "Received UL measurements contain %u SUB measurements, expected at least %d\n",
+			  num_meas_sub_actual, num_meas_sub_expect);
+		if (OSMO_UNLIKELY(num_meas_sub < num_meas_sub_expect)) {
+			LOGPLCHAN(lchan, DMEAS, LOGL_ERROR,
+				  "Incorrect number of SUB measurements detected! "
+				  "(%u vs exp >=%d)\n", num_meas_sub, num_meas_sub_expect);
+		}
+	} else {
+		LOGPLCHAN(lchan, DMEAS, LOGL_DEBUG,
+			  "Received UL measurements contain %u SUB measurements, expected %d\n",
+			  num_meas_sub_actual, num_meas_sub_expect);
+		if (OSMO_UNLIKELY(num_meas_sub != num_meas_sub_expect)) {
+			LOGPLCHAN(lchan, DMEAS, LOGL_ERROR,
+				  "Incorrect number of SUB measurements detected! "
+				  "(%u vs exp %d)\n", num_meas_sub, num_meas_sub_expect);
+		}
 	}
 
 	/* Measurement computation step 2: divide */
 	ber_full_sum = ber_full_sum / num_ul_meas;
 
 	if (!irssi_full_sum)
-		ber_full_sum = MEASUREMENT_DUMMY_IRSSI;
+		irssi_full_sum = MEASUREMENT_DUMMY_IRSSI;
 	else
 		irssi_full_sum = irssi_full_sum / num_ul_meas_actual;
 
-	if (!num_ul_meas_actual)
+	if (!num_ul_meas_actual) {
 		ta256b_sum = lchan->meas.ms_toa256;
-	else
-		ta256b_sum = ta256b_sum / num_ul_meas_actual;
+		ci_full_sum = lchan->meas.ul_ci_cb_full;
+	} else {
+		ta256b_sum = ta256b_sum / (signed)num_ul_meas_actual;
+		ci_full_sum = ci_full_sum / (signed)num_ul_meas_actual;
+	}
 
 	if (!num_meas_sub)
 		ber_sub_sum = MEASUREMENT_DUMMY_BER;
 	else
 		ber_sub_sum = ber_sub_sum / num_meas_sub;
 
-	if (!num_meas_sub_actual)
+	if (!num_meas_sub_actual) {
 		irssi_sub_sum = MEASUREMENT_DUMMY_IRSSI;
-	else
+		ci_sub_sum = lchan->meas.ul_ci_cb_sub;
+	} else {
 		irssi_sub_sum = irssi_sub_sum / num_meas_sub_actual;
+		ci_sub_sum = ci_sub_sum / (signed)num_meas_sub_actual;
+	}
 
-	LOGP(DMEAS, LOGL_INFO, "%s Computed TA256(% 4d) BER-FULL(%2u.%02u%%), RSSI-FULL(-%3udBm), "
-	     "BER-SUB(%2u.%02u%%), RSSI-SUB(-%3udBm)\n", gsm_lchan_name(lchan),
-	     ta256b_sum, ber_full_sum / 100,
-	     ber_full_sum % 100, irssi_full_sum, ber_sub_sum / 100, ber_sub_sum % 100, irssi_sub_sum);
+	LOGPLCHAN(lchan, DMEAS, LOGL_INFO,
+		  "Computed TA256(% 4d), BER-FULL(%2u.%02u%%), RSSI-FULL(-%3udBm), C/I-FULL(% 4d cB), "
+		  "BER-SUB(%2u.%02u%%), RSSI-SUB(-%3udBm), C/I-SUB(% 4d cB)\n",
+		  ta256b_sum, ber_full_sum / 100, ber_full_sum % 100, irssi_full_sum, ci_full_sum,
+		  ber_sub_sum / 100, ber_sub_sum % 100, irssi_sub_sum, ci_sub_sum);
 
 	/* store results */
 	mru = &lchan->meas.ul_res;
@@ -686,11 +742,15 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
 	mru->full.rx_qual = ber10k_to_rxqual(ber_full_sum);
 	mru->sub.rx_qual = ber10k_to_rxqual(ber_sub_sum);
 	lchan->meas.ms_toa256 = ta256b_sum;
+	lchan->meas.ul_ci_cb_full = ci_full_sum;
+	lchan->meas.ul_ci_cb_sub = ci_sub_sum;
 
-	LOGP(DMEAS, LOGL_INFO, "%s UL MEAS RXLEV_FULL(%u), RXLEV_SUB(%u),"
-	     "RXQUAL_FULL(%u), RXQUAL_SUB(%u), num_meas_sub(%u), num_ul_meas(%u) \n",
-	     gsm_lchan_name(lchan),
-	     mru->full.rx_lev, mru->sub.rx_lev, mru->full.rx_qual, mru->sub.rx_qual, num_meas_sub, num_ul_meas_expect);
+	LOGPLCHAN(lchan, DMEAS, LOGL_INFO,
+		  "UL MEAS RXLEV_FULL(%u), RXLEV_SUB(%u), RXQUAL_FULL(%u), RXQUAL_SUB(%u), "
+		  "num_meas_sub(%u), num_ul_meas(%u)\n",
+		  mru->full.rx_lev, mru->sub.rx_lev,
+		  mru->full.rx_qual, mru->sub.rx_qual,
+		  num_meas_sub, num_ul_meas_expect);
 
 	lchan->meas.flags |= LC_UL_M_F_RES_VALID;
 
@@ -707,7 +767,9 @@ int lchan_meas_check_compute(struct gsm_lchan *lchan, uint32_t fn)
  * l1sap.c every time a measurement indication is received. It collects the
  * measurement samples and automatically detects the end of the measurement
  * interval. */
-int lchan_meas_process_measurement(struct gsm_lchan *lchan, struct bts_ul_meas *ulm, uint32_t fn)
+int lchan_meas_process_measurement(struct gsm_lchan *lchan,
+				   const struct bts_ul_meas *ulm,
+				   uint32_t fn)
 {
 	lchan_new_ul_meas(lchan, ulm, fn);
 	return lchan_meas_check_compute(lchan, fn);
@@ -721,3 +783,230 @@ void lchan_meas_reset(struct gsm_lchan *lchan)
 	memset(&lchan->meas, 0, sizeof(lchan->meas));
 	lchan->meas.last_fn = LCHAN_FN_DUMMY;
 }
+
+static inline uint8_t ms_to2rsl(const struct gsm_lchan *lchan, uint8_t ta)
+{
+	return (lchan->ms_t_offs >= 0) ? lchan->ms_t_offs : (lchan->p_offs - ta);
+}
+
+static inline bool ms_to_valid(const struct gsm_lchan *lchan)
+{
+	return (lchan->ms_t_offs >= 0) || (lchan->p_offs >= 0);
+}
+
+/* Decide if repeated FACCH should be applied or not. If RXQUAL level, that the
+ * MS reports is high enough, FACCH repetition is not needed. */
+static void repeated_dl_facch_active_decision(struct gsm_lchan *lchan,
+					      const struct gsm48_meas_res *meas_res)
+{
+	uint8_t upper;
+	uint8_t lower;
+	uint8_t rxqual;
+	bool prev_repeated_dl_facch_active = lchan->rep_acch.dl_facch_active;
+
+	/* This is an optimization so that we exit as quickly as possible if
+	 * there are no FACCH repetition capabilities present. However If the
+	 * repeated FACCH capabilities vanish for whatever reason, we must be
+	 * sure that FACCH repetition is disabled. */
+	if (!lchan->rep_acch_cap.dl_facch_cmd
+	    && !lchan->rep_acch_cap.dl_facch_all) {
+		lchan->rep_acch.dl_facch_active = false;
+		goto out;
+	}
+
+	/* Threshold disabled (always on) */
+	if (lchan->rep_acch_cap.rxqual == 0) {
+		lchan->rep_acch.dl_facch_active = true;
+		goto out;
+	}
+
+	/* When the MS sets the SRR bit in the UL-SACCH L1 header
+	 * (repeated SACCH requested) then it makes sense to enable
+	 * FACCH repetition too. */
+	if (lchan->meas.l1_info.srr_sro) {
+		lchan->rep_acch.dl_facch_active = true;
+		goto out;
+	}
+
+	/* Parse MS measurement results */
+	if (meas_res == NULL)
+		goto out;
+	if (!gsm48_meas_res_is_valid(meas_res))
+		goto out;
+
+	/* If the RXQUAL level at the MS drops under a certain threshold
+	 * we enable FACCH repetition. */
+	upper = lchan->rep_acch_cap.rxqual;
+	if (upper > 2)
+		lower = lchan->rep_acch_cap.rxqual - 2;
+	else
+		lower = 0;
+
+	/* When downlink DTX is applied, use RXQUAL-SUB, otherwise use
+	 * RXQUAL-FULL. */
+	if (meas_res->dtx_used)
+		rxqual = meas_res->rxqual_sub;
+	else
+		rxqual = meas_res->rxqual_full;
+
+	if (rxqual >= upper)
+		lchan->rep_acch.dl_facch_active = true;
+	else if (rxqual <= lower)
+		lchan->rep_acch.dl_facch_active = false;
+
+out:
+	if (lchan->rep_acch.dl_facch_active == prev_repeated_dl_facch_active)
+		return;
+	if (lchan->rep_acch.dl_facch_active)
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-FACCH repetition: inactive => active\n");
+	else
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "DL-FACCH repetition: active => inactive\n");
+}
+
+static void acch_overpower_active_decision(struct gsm_lchan *lchan,
+					   const struct gsm48_meas_res *meas_res)
+{
+	const bool old = lchan->top_acch_active;
+	uint8_t upper, lower, rxqual;
+
+	/* ACCH overpower is not allowed => nothing to do */
+	if (lchan->top_acch_cap.overpower_db == 0)
+		return;
+	/* RxQual threshold is disabled => overpower is always on */
+	if (lchan->top_acch_cap.rxqual == 0)
+		return;
+
+	/* If DTx is active on Downlink, use the '-SUB' */
+	if (meas_res->dtx_used)
+		rxqual = meas_res->rxqual_sub;
+	else /* ... otherwise use the '-FULL' */
+		rxqual = meas_res->rxqual_full;
+
+	upper = lchan->top_acch_cap.rxqual;
+	if (upper > 2)
+		lower = upper - 2;
+	else
+		lower = 0;
+
+	if (rxqual >= upper)
+		lchan->top_acch_active = true;
+	else if (rxqual <= lower)
+		lchan->top_acch_active = false;
+
+	if (lchan->top_acch_active != old) {
+		LOGPLCHAN(lchan, DL1P, LOGL_DEBUG, "Temporary ACCH overpower: %s\n",
+			  lchan->top_acch_active ? "inactive => active"
+						  : "active => inactive");
+	}
+}
+
+static bool data_is_rr_meas_rep(const uint8_t *data)
+{
+	const struct gsm48_hdr *gh = (void *)(data + 5);
+	const uint8_t *lapdm_hdr = (void *)(data + 2);
+
+	/* LAPDm address field: SAPI=0, C/R=0, EA=1 */
+	if (lapdm_hdr[0] != 0x01)
+		return false;
+	/* LAPDm control field: U, func=UI */
+	if (lapdm_hdr[1] != 0x03)
+		return false;
+	/* Protocol discriminator: RR */
+	if (gh->proto_discr != GSM48_PDISC_RR)
+		return false;
+
+	switch (gh->msg_type) {
+	case GSM48_MT_RR_EXT_MEAS_REP:
+	case GSM48_MT_RR_MEAS_REP:
+		return true;
+	default:
+		return false;
+	}
+}
+
+/* Called every time a SACCH block is received from lower layers */
+void lchan_meas_handle_sacch(struct gsm_lchan *lchan, struct msgb *msg)
+{
+	const struct gsm48_meas_res *mr = NULL;
+	const struct gsm48_hdr *gh = NULL;
+	int timing_offset, rc;
+	bool dtxu_used = true; /* safe default assumption */
+	uint8_t ms_pwr;
+	uint8_t ms_ta;
+	int8_t ul_rssi;
+	int16_t ul_ci_cb;
+	uint8_t *l3;
+	unsigned int l3_len;
+
+	if (msgb_l2len(msg) == GSM_MACBLOCK_LEN) {
+		/* Some brilliant engineer decided that the ordering of
+		 * fields on the Um interface is different from the
+		 * order of fields in RSL. See 3GPP TS 44.004 (section 7.2)
+		 * vs. 3GPP TS 48.058 (section 9.3.10). */
+		const struct gsm_sacch_l1_hdr *l1h = msgb_l2(msg);
+		lchan->meas.l1_info.ms_pwr = l1h->ms_pwr;
+		lchan->meas.l1_info.fpc_epc = l1h->fpc_epc;
+		lchan->meas.l1_info.srr_sro = l1h->srr_sro;
+		lchan->meas.l1_info.ta = l1h->ta;
+		lchan->meas.flags |= LC_UL_M_F_L1_VALID;
+
+		/* Check if this is a Measurement Report */
+		if (data_is_rr_meas_rep(msgb_l2(msg))) {
+			/* Skip both L1 SACCH and LAPDm headers */
+			msg->l3h = (void *)(msg->l2h + 2 + 3);
+			gh = msgb_l3(msg);
+		}
+
+		ms_pwr = lchan->meas.l1_info.ms_pwr;
+		ms_ta = lchan->meas.l1_info.ta;
+	} else {
+		lchan->meas.flags &= ~LC_UL_M_F_L1_VALID;
+		ms_pwr = lchan->ms_power_ctrl.current;
+		ms_ta = lchan->ta_ctrl.current;
+	}
+
+	timing_offset = ms_to_valid(lchan) ? ms_to2rsl(lchan, ms_ta) : -1;
+	l3 = msgb_l3(msg);
+	l3_len = l3 ? msgb_l3len(msg) : 0;
+	rc = rsl_tx_meas_res(lchan, l3, l3_len, timing_offset);
+	if (rc == 0) /* Count successful transmissions */
+		lchan->meas.res_nr++;
+
+	/* Run control loops now that we have all the information: */
+	/* 3GPP TS 45.008 sec 4.2: UL L1 SACCH Header contains TA and
+	 * MS_PWR used "for the last burst of the previous SACCH
+	 * period". Since MS must use the values provided in DL SACCH
+	 * starting at next meas period, the value of the "last burst"
+	 * is actually the value used in the entire meas period. Since
+	 * it contains info about the previous meas period, we want to
+	 * feed the Control Loop with the measurements for the same
+	 * period (the previous one), which is stored in lchan->meas(.ul_res):
+	 */
+	if (gh && gh->msg_type == GSM48_MT_RR_MEAS_REP) {
+		mr = (const struct gsm48_meas_res *)gh->data;
+		if (gsm48_meas_res_is_valid(mr))
+			dtxu_used = mr->dtx_used;
+	}
+
+	if (dtxu_used) {
+		ul_rssi = rxlev2dbm(lchan->meas.ul_res.sub.rx_lev);
+		ul_ci_cb = lchan->meas.ul_ci_cb_sub;
+	} else {
+		ul_rssi = rxlev2dbm(lchan->meas.ul_res.full.rx_lev);
+		ul_ci_cb = lchan->meas.ul_ci_cb_full;
+	}
+	lchan_ms_ta_ctrl(lchan, ms_ta, lchan->meas.ms_toa256);
+	lchan_ms_pwr_ctrl(lchan, ms_pwr, ul_rssi, ul_ci_cb);
+	if (mr && gsm48_meas_res_is_valid(mr)) {
+		lchan_bs_pwr_ctrl(lchan, mr);
+		acch_overpower_active_decision(lchan, mr);
+	}
+
+	repeated_dl_facch_active_decision(lchan, mr);
+
+	/* Reset state for next iteration */
+	lchan->tch.dtx.dl_active = false;
+	lchan->meas.flags &= ~LC_UL_M_F_OSMO_EXT_VALID;
+	lchan->ms_t_offs = -1;
+	lchan->p_offs = -1;
+}
