@@ -15,7 +15,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -35,6 +35,7 @@
 #include <osmocom/core/stats.h>
 
 #include <osmocom/gsm/protocol/gsm_08_58.h>
+#include <osmocom/gsm/gsm0502.h>
 #include <osmocom/gsm/a5.h>
 
 #include <osmo-bts/gsm_data.h>
@@ -150,7 +151,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.desc = "Frequency correction channel",
 
 		/* Tx only, frequency correction bursts */
-		.flags = TRX_CHAN_FLAG_AUTO_ACTIVE,
 		.dl_fn = tx_fcch_fn,
 	},
 	[TRXC_SCH] = {
@@ -158,7 +158,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.desc = "Synchronization channel",
 
 		/* Tx only, synchronization bursts */
-		.flags = TRX_CHAN_FLAG_AUTO_ACTIVE,
 		.dl_fn = tx_sch_fn,
 	},
 	[TRXC_BCCH] = {
@@ -169,7 +168,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		/* Tx only, xCCH convolutional coding (3GPP TS 05.03, section 4.4),
 		 * regular interleaving (3GPP TS 05.02, clause 7, table 3):
 		 * a L2 frame is interleaved over 4 consecutive bursts. */
-		.flags = TRX_CHAN_FLAG_AUTO_ACTIVE,
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 	},
@@ -179,7 +177,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.chan_nr = RSL_CHAN_RACH,
 
 		/* Rx only, RACH convolutional coding (3GPP TS 05.03, section 4.6). */
-		.flags = TRX_CHAN_FLAG_AUTO_ACTIVE,
 		.ul_fn = rx_rach_fn,
 	},
 	[TRXC_CCCH] = {
@@ -190,7 +187,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		/* Tx only, xCCH convolutional coding (3GPP TS 05.03, section 4.4),
 		 * regular interleaving (3GPP TS 05.02, clause 7, table 3):
 		 * a L2 frame is interleaved over 4 consecutive bursts. */
-		.flags = TRX_CHAN_FLAG_AUTO_ACTIVE,
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 	},
@@ -547,9 +543,7 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.chan_nr = RSL_CHAN_OSMO_PDCH,
 
 		/* Rx and Tx, multiple coding schemes: CS-2..4 and MCS-1..9 (3GPP TS
-		 * 05.03, chapter 5), regular interleaving as specified for xCCH.
-		 * NOTE: the burst buffer is three times bigger because the
-		 * payload of EDGE bursts is three times longer. */
+		 * 05.03, chapter 5), regular interleaving as specified for xCCH. */
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_pdtch_fn,
 		.ul_fn = rx_pdtch_fn,
@@ -575,7 +569,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.chan_nr = RSL_CHAN_OSMO_CBCH4,
 
 		/* Tx only, same as for TRXC_BCCH (xCCH), see above. */
-		.flags = TRX_CHAN_FLAG_AUTO_ACTIVE,
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 	},
@@ -659,7 +652,7 @@ static void trx_sched_clean_ts(struct gsm_bts_trx_ts *ts)
 	struct l1sched_ts *l1ts = ts->priv;
 	unsigned int i;
 
-	msgb_queue_flush(&l1ts->dl_prims);
+	msgb_queue_free(&l1ts->dl_prims);
 	rate_ctr_group_free(l1ts->ctrs);
 	l1ts->ctrs = NULL;
 
@@ -757,10 +750,10 @@ free_msg:
 }
 
 int _sched_compose_ph_data_ind(struct l1sched_ts *l1ts, uint32_t fn,
-			       enum trx_chan_type chan, uint8_t *l2,
-			       uint8_t l2_len, float rssi,
+			       enum trx_chan_type chan,
+			       const uint8_t *data, size_t data_len,
+			       uint16_t ber10k, float rssi,
 			       int16_t ta_offs_256bits, int16_t link_qual_cb,
-			       uint16_t ber10k,
 			       enum osmo_ph_pres_info_type presence_info)
 {
 	struct msgb *msg;
@@ -772,7 +765,7 @@ int _sched_compose_ph_data_ind(struct l1sched_ts *l1ts, uint32_t fn,
 		chan_nr |= RSL_CHAN_OSMO_VAMOS_MASK;
 
 	/* compose primitive */
-	msg = l1sap_msgb_alloc(l2_len);
+	msg = l1sap_msgb_alloc(data_len);
 	l1sap = msgb_l1sap_prim(msg);
 	osmo_prim_init(&l1sap->oph, SAP_GSM_PH, PRIM_PH_DATA,
 		PRIM_OP_INDICATION, msg);
@@ -784,9 +777,9 @@ int _sched_compose_ph_data_ind(struct l1sched_ts *l1ts, uint32_t fn,
 	l1sap->u.data.ta_offs_256bits = ta_offs_256bits;
 	l1sap->u.data.lqual_cb = link_qual_cb;
 	l1sap->u.data.pdch_presence_info = presence_info;
-	msg->l2h = msgb_put(msg, l2_len);
-	if (l2_len)
-		memcpy(msg->l2h, l2, l2_len);
+	msg->l2h = msgb_put(msg, data_len);
+	if (data_len)
+		memcpy(msg->l2h, data, data_len);
 
 	/* forward primitive */
 	l1sap_up(l1ts->ts->trx, l1sap);
@@ -795,9 +788,11 @@ int _sched_compose_ph_data_ind(struct l1sched_ts *l1ts, uint32_t fn,
 }
 
 int _sched_compose_tch_ind(struct l1sched_ts *l1ts, uint32_t fn,
-			   enum trx_chan_type chan, uint8_t *tch, uint8_t tch_len,
-			   int16_t ta_offs_256bits, uint16_t ber10k, float rssi,
-			   int16_t link_qual_cb, uint8_t is_sub)
+			   enum trx_chan_type chan,
+			   const uint8_t *data, size_t data_len,
+			   uint16_t ber10k, float rssi,
+			   int16_t ta_offs_256bits, int16_t link_qual_cb,
+			   uint8_t is_sub)
 {
 	struct msgb *msg;
 	struct osmo_phsap_prim *l1sap;
@@ -809,7 +804,7 @@ int _sched_compose_tch_ind(struct l1sched_ts *l1ts, uint32_t fn,
 		chan_nr |= RSL_CHAN_OSMO_VAMOS_MASK;
 
 	/* compose primitive */
-	msg = l1sap_msgb_alloc(tch_len);
+	msg = l1sap_msgb_alloc(data_len);
 	l1sap = msgb_l1sap_prim(msg);
 	osmo_prim_init(&l1sap->oph, SAP_GSM_PH, PRIM_TCH,
 		PRIM_OP_INDICATION, msg);
@@ -821,12 +816,12 @@ int _sched_compose_tch_ind(struct l1sched_ts *l1ts, uint32_t fn,
 	l1sap->u.tch.lqual_cb = link_qual_cb;
 	l1sap->u.tch.is_sub = is_sub & 1;
 
-	msg->l2h = msgb_put(msg, tch_len);
-	if (tch_len)
-		memcpy(msg->l2h, tch, tch_len);
+	msg->l2h = msgb_put(msg, data_len);
+	if (data_len)
+		memcpy(msg->l2h, data, data_len);
 
-	LOGL1S(DL1P, LOGL_DEBUG, l1ts, chan, l1sap->u.data.fn, "%s Rx -> RTP: %s\n",
-	       gsm_lchan_name(lchan), osmo_hexdump(msgb_l2(msg), msgb_l2len(msg)));
+	LOGL1S(DL1P, LOGL_DEBUG, l1ts, chan, l1sap->u.tch.fn, "%s Rx -> RTP: %s\n",
+	       gsm_lchan_name(lchan), msgb_hexdump_l2(msg));
 	/* forward primitive */
 	l1sap_up(l1ts->ts->trx, l1sap);
 
@@ -1013,8 +1008,8 @@ int trx_sched_set_pchan(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config pch
 	struct l1sched_ts *l1ts = ts->priv;
 	int i = find_sched_mframe_idx(pchan, ts->nr);
 	if (i < 0) {
-		LOGP(DL1C, LOGL_NOTICE, "%s Failed to configure multiframe\n",
-		     gsm_ts_name(ts));
+		LOGP(DL1C, LOGL_NOTICE, "%s Failed to configure multiframe (pchan=0x%02x)\n",
+		     gsm_ts_name(ts), pchan);
 		return -ENOTSUP;
 	}
 	l1ts->mf_index = i;
@@ -1062,6 +1057,51 @@ static void trx_sched_queue_filter(struct llist_head *q, uint8_t chan_nr, uint8_
 	}
 }
 
+static void _trx_sched_set_lchan(struct gsm_lchan *lchan,
+				 enum trx_chan_type chan,
+				 bool active)
+{
+	struct l1sched_ts *l1ts = lchan->ts->priv;
+	struct l1sched_chan_state *chan_state;
+
+	OSMO_ASSERT(l1ts != NULL);
+	chan_state = &l1ts->chan_state[chan];
+
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "%s %s\n",
+		  (active) ? "Activating" : "Deactivating",
+		  trx_chan_desc[chan].name);
+
+	if (active) {
+		/* Clean up everything */
+		memset(chan_state, 0, sizeof(*chan_state));
+
+		/* Bind to generic 'struct gsm_lchan' */
+		chan_state->lchan = lchan;
+
+		/* Allocate memory for Rx/Tx burst buffers.  Use the maximim size
+		 * of 24 * (2 * 58) bytes, which is sufficient to store up to 24 GMSK
+		 * modulated bursts for CSD or up to 8 8PSK modulated bursts for EGPRS. */
+		const size_t buf_size = 24 * GSM_NBITS_NB_GMSK_PAYLOAD;
+		if (trx_chan_desc[chan].dl_fn != NULL)
+			chan_state->dl_bursts = talloc_zero_size(l1ts, buf_size);
+		if (trx_chan_desc[chan].ul_fn != NULL)
+			chan_state->ul_bursts = talloc_zero_size(l1ts, buf_size);
+	} else {
+		chan_state->ho_rach_detect = 0;
+
+		/* Remove pending Tx prims belonging to this lchan */
+		trx_sched_queue_filter(&l1ts->dl_prims,
+				       trx_chan_desc[chan].chan_nr,
+				       trx_chan_desc[chan].link_id);
+
+		/* Release memory used by Rx/Tx burst buffers */
+		TALLOC_FREE(chan_state->dl_bursts);
+		TALLOC_FREE(chan_state->ul_bursts);
+	}
+
+	chan_state->active = active;
+}
+
 /* setting all logical channels given attributes to active/inactive */
 int trx_sched_set_lchan(struct gsm_lchan *lchan, uint8_t chan_nr, uint8_t link_id, bool active)
 {
@@ -1069,7 +1109,6 @@ int trx_sched_set_lchan(struct gsm_lchan *lchan, uint8_t chan_nr, uint8_t link_i
 	uint8_t tn = L1SAP_CHAN2TS(chan_nr);
 	uint8_t ss = l1sap_chan2ss(chan_nr);
 	bool found = false;
-	int i;
 
 	if (!l1ts) {
 		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "%s lchan with uninitialized scheduler structure\n",
@@ -1083,48 +1122,15 @@ int trx_sched_set_lchan(struct gsm_lchan *lchan, uint8_t chan_nr, uint8_t link_i
 		chan_nr &= ~RSL_CHAN_OSMO_VAMOS_MASK;
 
 	/* look for all matching chan_nr/link_id */
-	for (i = 0; i < _TRX_CHAN_MAX; i++) {
-		struct l1sched_chan_state *chan_state = &l1ts->chan_state[i];
-
-		if (trx_chan_desc[i].chan_nr != (chan_nr & RSL_CHAN_NR_MASK))
+	for (enum trx_chan_type chan = 0; chan < _TRX_CHAN_MAX; chan++) {
+		if (trx_chan_desc[chan].chan_nr != (chan_nr & RSL_CHAN_NR_MASK))
 			continue;
-		if (trx_chan_desc[i].link_id != link_id)
+		if (trx_chan_desc[chan].link_id != link_id)
 			continue;
-		if (chan_state->active == active)
+		if (l1ts->chan_state[chan].active == active)
 			continue;
 		found = true;
-
-		LOGPLCHAN(lchan, DL1C, LOGL_INFO, "%s %s\n",
-			  (active) ? "Activating" : "Deactivating",
-			  trx_chan_desc[i].name);
-		/* free burst memory, to cleanly start with burst 0 */
-		if (chan_state->dl_bursts) {
-			talloc_free(chan_state->dl_bursts);
-			chan_state->dl_bursts = NULL;
-		}
-		if (chan_state->ul_bursts) {
-			talloc_free(chan_state->ul_bursts);
-			chan_state->ul_bursts = NULL;
-		}
-		if (chan_state->ul_bursts_prev) {
-			talloc_free(chan_state->ul_bursts_prev);
-			chan_state->ul_bursts_prev = NULL;
-		}
-
-		if (active) {
-			/* Clean up everything */
-			memset(chan_state, 0, sizeof(*chan_state));
-
-			/* Bind to generic 'struct gsm_lchan' */
-			chan_state->lchan = lchan;
-		} else {
-			chan_state->ho_rach_detect = 0;
-
-			/* Remove pending Tx prims belonging to this lchan */
-			trx_sched_queue_filter(&l1ts->dl_prims, chan_nr, link_id);
-		}
-
-		chan_state->active = active;
+		_trx_sched_set_lchan(lchan, chan, active);
 	}
 
 	/* disable handover detection (on deactivation) */
@@ -1134,6 +1140,58 @@ int trx_sched_set_lchan(struct gsm_lchan *lchan, uint8_t chan_nr, uint8_t link_i
 	return found ? 0 : -EINVAL;
 }
 
+int trx_sched_set_ul_access(struct gsm_lchan *lchan, uint8_t chan_nr, bool active)
+{
+	struct l1sched_ts *l1ts = lchan->ts->priv;
+	uint8_t tn = L1SAP_CHAN2TS(chan_nr);
+	uint8_t ss = l1sap_chan2ss(chan_nr);
+	int i;
+
+	if (!l1ts) {
+		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "%s UL access on lchan with uninitialized scheduler structure.\n",
+			  (active) ? "Activating" : "Deactivating");
+		return -EINVAL;
+	}
+
+	/* look for all matching chan_nr */
+	for (i = 0; i < _TRX_CHAN_MAX; i++) {
+		if (trx_chan_desc[i].chan_nr == (chan_nr & RSL_CHAN_NR_MASK)) {
+			struct l1sched_chan_state *l1cs = &l1ts->chan_state[i];
+
+			l1cs->ho_rach_detect = active;
+		}
+	}
+
+	_sched_act_rach_det(lchan->ts->trx, tn, ss, active);
+
+	return 0;
+}
+
+int trx_sched_set_bcch_ccch(struct gsm_lchan *lchan, bool active)
+{
+	struct l1sched_ts *l1ts = lchan->ts->priv;
+	static const enum trx_chan_type chans[] = {
+		TRXC_FCCH,
+		TRXC_SCH,
+		TRXC_BCCH,
+		TRXC_RACH,
+		TRXC_CCCH,
+	};
+
+	if (!l1ts)
+		return -EINVAL;
+
+	for (unsigned int i = 0; i < ARRAY_SIZE(chans); i++) {
+		enum trx_chan_type chan = chans[i];
+
+		if (l1ts->chan_state[chan].active == active)
+			continue;
+		_trx_sched_set_lchan(lchan, chan, active);
+	}
+
+	return 0;
+}
+
 /* setting all logical channels given attributes to active/inactive */
 int trx_sched_set_mode(struct gsm_bts_trx_ts *ts, uint8_t chan_nr, uint8_t rsl_cmode,
 	uint8_t tch_mode, int codecs, uint8_t codec0, uint8_t codec1,
@@ -1277,7 +1335,7 @@ int _sched_rts(const struct l1sched_ts *l1ts, uint32_t fn)
 		return 0;
 
 	/* check if channel is active */
-	if (!TRX_CHAN_IS_ACTIVE(&l1ts->chan_state[chan], chan))
+	if (!l1ts->chan_state[chan].active)
 	 	return -EINVAL;
 
 	/* There is no burst, just for logging */
@@ -1334,7 +1392,7 @@ void _sched_dl_burst(struct l1sched_ts *l1ts, struct trx_dl_burst_req *br)
 	l1cs = &l1ts->chan_state[br->chan];
 
 	/* check if channel is active */
-	if (!TRX_CHAN_IS_ACTIVE(l1cs, br->chan))
+	if (!l1cs->active)
 		return;
 
 	/* Training Sequence Code and Set */
@@ -1523,7 +1581,7 @@ int trx_sched_ul_burst(struct l1sched_ts *l1ts, struct trx_ul_burst_ind *bi)
 	func = trx_chan_desc[bi->chan].ul_fn;
 
 	/* check if channel is active */
-	if (!TRX_CHAN_IS_ACTIVE(l1cs, bi->chan)) {
+	if (!l1cs->active) {
 		/* handle noise measurements on dedicated and idle channels */
 		if (TRX_CHAN_IS_DEDIC(bi->chan) || bi->chan == TRXC_IDLE)
 			trx_sched_noise_meas(l1cs, bi);
