@@ -12,7 +12,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -22,6 +22,7 @@
 #include "btsconfig.h"
 
 #include <inttypes.h>
+#include <limits.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -43,12 +44,14 @@
 #include <osmocom/core/sockaddr_str.h>
 #include <osmocom/trau/osmo_ortp.h>
 #include <osmocom/core/fsm.h>
+#include <osmocom/codec/codec.h>
 
 #include <osmo-bts/logging.h>
 #include <osmo-bts/gsm_data.h>
 #include <osmo-bts/phy_link.h>
 #include <osmo-bts/abis.h>
 #include <osmo-bts/bts.h>
+#include <osmo-bts/bts_sm.h>
 #include <osmo-bts/rsl.h>
 #include <osmo-bts/oml.h>
 #include <osmo-bts/signal.h>
@@ -57,6 +60,7 @@
 #include <osmo-bts/measurement.h>
 #include <osmo-bts/vty.h>
 #include <osmo-bts/l1sap.h>
+#include <osmo-bts/osmux.h>
 
 #define VTY_STR	"Configure the VTY\n"
 
@@ -67,6 +71,9 @@
 #define BTS_TRX_STR BTS_NR_STR TRX_NR_STR
 #define BTS_TRX_TS_STR BTS_TRX_STR TS_NR_STR
 #define BTS_TRX_TS_LCHAN_STR BTS_TRX_TS_STR LCHAN_NR_STR
+/* INT32_MAX, because osmo_wqueue_init takes int as an argument
+ * and INT_MAX can't be stringified as a decimal */
+#define BTS_CFG_PCU_SOCK_WQUEUE_LEN_MAX_MAX 2147483647
 
 #define X(x) (1 << x)
 
@@ -133,21 +140,9 @@ int bts_vty_go_parent(struct vty *vty)
 	return vty->node;
 }
 
-int bts_vty_is_config_node(struct vty *vty, int node)
-{
-	switch (node) {
-	case TRX_NODE:
-	case BTS_NODE:
-	case PHY_NODE:
-	case PHY_INST_NODE:
-		return 1;
-	default:
-		return 0;
-	}
-}
-
 static const char osmobts_copyright[] =
-	"Copyright (C) 2010, 2011 by Harald Welte, Andreas Eversberg and On-Waves\r\n"
+	"Copyright (C) 2010-2011 by Harald Welte, Andreas Eversberg and On-Waves\r\n"
+	"Copyright (C) 2011-2022 by sysmocom - s.f.m.c. GmbH\r\n"
 	"License AGPLv3+: GNU AGPL version 3 or later <http://gnu.org/licenses/agpl-3.0.html>\r\n"
 	"This is free software: you are free to change and redistribute it.\r\n"
 	 "There is NO WARRANTY, to the extent permitted by law.\r\n";
@@ -157,7 +152,6 @@ struct vty_app_info bts_vty_info = {
 	.version	= PACKAGE_VERSION,
 	.copyright	= osmobts_copyright,
 	.go_parent_cb	= bts_vty_go_parent,
-	.is_config_node	= bts_vty_is_config_node,
 	.usr_attr_desc	= {
 		[BTS_VTY_ATTR_NEW_LCHAN] = \
 			"This command applies for newly created lchans",
@@ -170,13 +164,6 @@ struct vty_app_info bts_vty_info = {
 	},
 };
 
-extern struct gsm_network bts_gsmnet;
-
-struct gsm_network *gsmnet_from_vty(struct vty *v)
-{
-	return &bts_gsmnet;
-}
-
 static struct cmd_node bts_node = {
 	BTS_NODE,
 	"%s(bts)# ",
@@ -189,6 +176,12 @@ static struct cmd_node trx_node = {
 	1,
 };
 
+static struct cmd_node osmux_node = {
+	OSMUX_NODE,
+	"%s(osmux)# ",
+	1,
+};
+
 gDEFUN(cfg_bts_auto_band, cfg_bts_auto_band_cmd,
 	"auto-band",
 	"Automatically select band for ARFCN based on configured band\n")
@@ -209,6 +202,90 @@ gDEFUN(cfg_bts_no_auto_band, cfg_bts_no_auto_band_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN_ATTR(cfg_bts_osmux, cfg_bts_osmux_cmd,
+	   "osmux",
+	   "Configure Osmux\n",
+	   CMD_ATTR_IMMEDIATE)
+{
+	vty->node = OSMUX_NODE;
+	return CMD_SUCCESS;
+}
+
+DEFUN_ATTR(cfg_bts_osmux_use, cfg_bts_osmux_use_cmd,
+	   "use (off|on|only)",
+	   "Configure Osmux usage\n"
+	   "Never use Osmux\n"
+	   "Use Osmux if requested by BSC (default)\n"
+	   "Always use Osmux, reject non-Osmux BSC requests\n",
+	   CMD_ATTR_IMMEDIATE)
+{
+	struct gsm_bts *bts = vty->index;
+	if (strcmp(argv[0], "off") == 0)
+		bts->osmux.use = OSMUX_USAGE_OFF;
+	else if (strcmp(argv[0], "on") == 0)
+		bts->osmux.use = OSMUX_USAGE_ON;
+	else if (strcmp(argv[0], "only") == 0)
+		bts->osmux.use = OSMUX_USAGE_ONLY;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_osmux_ip,
+      cfg_bts_osmux_ip_cmd,
+      "local-ip " VTY_IPV46_CMD,
+      IP_STR
+      "IPv4 Address to bind to\n"
+      "IPv6 Address to bind to\n")
+{
+	struct gsm_bts *bts = vty->index;
+	osmo_talloc_replace_string(bts, &bts->osmux.local_addr, argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_osmux_port,
+      cfg_bts_osmux_port_cmd,
+      "local-port <1-65535>",
+      "Osmux port\n" "UDP port\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->osmux.local_port = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_osmux_batch_factor,
+      cfg_bts_osmux_batch_factor_cmd,
+      "batch-factor <1-8>",
+      "Batching factor\n" "Number of messages in the batch\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->osmux.batch_factor = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_osmux_batch_size,
+      cfg_bts_osmux_batch_size_cmd,
+      "batch-size <1-65535>",
+      "Batch size\n" "Batch size in bytes\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->osmux.batch_size = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_osmux_dummy_padding,
+      cfg_bts_osmux_dummy_padding_cmd,
+      "dummy-padding (on|off)",
+      "Dummy padding\n"
+      "Enable dummy padding\n"
+      "Disable dummy padding (default)\n")
+{
+	struct gsm_bts *bts = vty->index;
+	if (strcmp(argv[0], "on") == 0)
+		bts->osmux.dummy_padding = true;
+	else if (strcmp(argv[0], "off") == 0)
+		bts->osmux.dummy_padding = false;
+	return CMD_SUCCESS;
+}
+
 DEFUN_ATTR(cfg_bts_trx, cfg_bts_trx_cmd,
 	   "trx <0-254>",
 	   "Select a TRX to configure\n" "TRX number\n",
@@ -278,6 +355,29 @@ static void config_write_dpc_params(struct vty *vty, const char *prefix,
 	}
 }
 
+static void config_write_osmux(struct vty *vty, const char *prefix, const struct gsm_bts *bts)
+{
+	vty_out(vty, "%sosmux%s", prefix, VTY_NEWLINE);
+	vty_out(vty, "%s use ", prefix);
+	switch (bts->osmux.use) {
+	case OSMUX_USAGE_ON:
+		vty_out(vty, "on%s", VTY_NEWLINE);
+		break;
+	case OSMUX_USAGE_ONLY:
+		vty_out(vty, "only%s", VTY_NEWLINE);
+		break;
+	case OSMUX_USAGE_OFF:
+	default:
+		vty_out(vty, "off%s", VTY_NEWLINE);
+		break;
+	}
+	vty_out(vty, "%s local-ip %s%s", prefix, bts->osmux.local_addr, VTY_NEWLINE);
+	vty_out(vty, "%s local-port %u%s", prefix, bts->osmux.local_port, VTY_NEWLINE);
+	vty_out(vty, "%s batch-factor %d%s", prefix, bts->osmux.batch_factor, VTY_NEWLINE);
+	vty_out(vty, "%s batch-size %u%s", prefix, bts->osmux.batch_size, VTY_NEWLINE);
+	vty_out(vty, "%s dummy-padding %s%s", prefix, bts->osmux.dummy_padding ? "on" : "off", VTY_NEWLINE);
+}
+
 static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 {
 	const struct gsm_bts_trx *trx;
@@ -302,9 +402,15 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 	vty_out(vty, " rtp port-range %u %u%s", bts->rtp_port_range_start,
 		bts->rtp_port_range_end, VTY_NEWLINE);
 	if (bts->rtp_ip_dscp != -1)
-		vty_out(vty, " rtp ip-dscp %i%s", bts->rtp_ip_dscp, VTY_NEWLINE);
+		vty_out(vty, " rtp ip-dscp %d%s", bts->rtp_ip_dscp, VTY_NEWLINE);
 	if (bts->rtp_priority != -1)
-		vty_out(vty, " rtp socket-priority %i%s", bts->rtp_priority, VTY_NEWLINE);
+		vty_out(vty, " rtp socket-priority %d%s", bts->rtp_priority, VTY_NEWLINE);
+	if (bts->rtp_nogaps_mode)
+		vty_out(vty, " rtp continuous-streaming%s", VTY_NEWLINE);
+	vty_out(vty, " %srtp internal-uplink-ecu%s",
+		bts->use_ul_ecu ? "" : "no ", VTY_NEWLINE);
+	vty_out(vty, " rtp hr-format %s%s",
+		bts->emit_hr_rfc5993 ? "rfc5993" : "ts101318", VTY_NEWLINE);
 	vty_out(vty, " paging queue-size %u%s", paging_get_queue_max(bts->paging_state),
 		VTY_NEWLINE);
 	vty_out(vty, " paging lifetime %u%s", paging_get_lifetime(bts->paging_state),
@@ -324,6 +430,10 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		vty_out(vty, " gsmtap-remote-host %s%s",
 			bts->gsmtap.remote_host,
 			VTY_NEWLINE);
+	if (bts->gsmtap.local_host != NULL)
+		vty_out(vty, " gsmtap-local-host %s%s",
+			bts->gsmtap.local_host,
+			VTY_NEWLINE);
 	for (i = 0; i < sizeof(uint32_t) * 8; i++) {
 		if (bts->gsmtap.sapi_mask & ((uint32_t) 1 << i)) {
 			sapi_buf = get_value_string_or_null(gsmtap_sapi_names, i);
@@ -337,6 +447,12 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		sapi_buf = osmo_str_tolower(get_value_string(gsmtap_sapi_names, GSMTAP_CHANNEL_ACCH));
 		vty_out(vty, " gsmtap-sapi %s%s", sapi_buf, VTY_NEWLINE);
 	}
+	if (bts->gsmtap.rlp) {
+		if (bts->gsmtap.rlp_skip_null)
+			vty_out(vty, " gsmtap-rlp skip-null%s", VTY_NEWLINE);
+		else
+			vty_out(vty, " gsmtap-rlp%s", VTY_NEWLINE);
+	}
 	vty_out(vty, " min-qual-rach %d%s", bts->min_qual_rach,
 		VTY_NEWLINE);
 	vty_out(vty, " min-qual-norm %d%s", bts->min_qual_norm,
@@ -345,12 +461,16 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		VTY_NEWLINE);
 	if (strcmp(bts->pcu.sock_path, PCU_SOCK_DEFAULT))
 		vty_out(vty, " pcu-socket %s%s", bts->pcu.sock_path, VTY_NEWLINE);
+	if (bts->pcu.sock_wqueue_len_max != BTS_CFG_PCU_SOCK_WQUEUE_LEN_MAX_MAX)
+		vty_out(vty, " pcu-socket-wqueue-length %u%s", bts->pcu.sock_wqueue_len_max, VTY_NEWLINE);
 	if (bts->supp_meas_toa256)
 		vty_out(vty, " supp-meas-info toa256%s", VTY_NEWLINE);
 	vty_out(vty, " smscb queue-max-length %d%s", bts->smscb_queue_max_len, VTY_NEWLINE);
 	vty_out(vty, " smscb queue-target-length %d%s", bts->smscb_queue_tgt_len, VTY_NEWLINE);
 	vty_out(vty, " smscb queue-hysteresis %d%s", bts->smscb_queue_hyst, VTY_NEWLINE);
 
+	config_write_osmux(vty, " ", bts);
+
 	bts_model_config_write_bts(vty, bts);
 
 	llist_for_each_entry(trx, &bts->trx_list, list) {
@@ -358,7 +478,7 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		const struct phy_instance *pinst = trx_phy_instance(trx);
 		vty_out(vty, " trx %u%s", trx->nr, VTY_NEWLINE);
 
-		if (trx->power_params.user_gain_mdB)
+		if (tpp->user_gain_mdB)
 			vty_out(vty, "  user-gain %u mdB%s",
 				tpp->user_gain_mdB, VTY_NEWLINE);
 		vty_out(vty, "  power-ramp max-initial %d mdBm%s",
@@ -381,10 +501,9 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 
 static int config_write_bts(struct vty *vty)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts;
 
-	llist_for_each_entry(bts, &net->bts_list, list)
+	llist_for_each_entry(bts, &g_bts_sm->bts_list, list)
 		config_write_bts_single(vty, bts);
 
 	osmo_tdef_vty_groups_write(vty, "");
@@ -443,16 +562,15 @@ DEFUN_ATTR(cfg_bts,
 	   "BTS Number\n",
 	   CMD_ATTR_IMMEDIATE)
 {
-	struct gsm_network *gsmnet = gsmnet_from_vty(vty);
 	int bts_nr = atoi(argv[0]);
 	struct gsm_bts *bts;
 
-	if (bts_nr >= gsmnet->num_bts) {
+	if (bts_nr >= g_bts_sm->num_bts) {
 		vty_out(vty, "%% Unknown BTS number %u (num %u)%s",
-			bts_nr, gsmnet->num_bts, VTY_NEWLINE);
+			bts_nr, g_bts_sm->num_bts, VTY_NEWLINE);
 		return CMD_WARNING;
 	} else
-		bts = gsm_bts_num(gsmnet, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 
 	vty->index = bts;
 	vty->index_sub = &bts->description;
@@ -658,6 +776,62 @@ DEFUN_USRATTR(cfg_bts_rtp_priority,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_rtp_cont_stream,
+      cfg_bts_rtp_cont_stream_cmd,
+      "rtp continuous-streaming",
+      RTP_STR "Always emit an RTP packet every 20 ms\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->rtp_nogaps_mode = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_no_rtp_cont_stream,
+      cfg_bts_no_rtp_cont_stream_cmd,
+      "no rtp continuous-streaming",
+      NO_STR RTP_STR "Always emit an RTP packet every 20 ms\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->rtp_nogaps_mode = false;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_rtp_int_ul_ecu,
+      cfg_bts_rtp_int_ul_ecu_cmd,
+      "rtp internal-uplink-ecu",
+      RTP_STR "Apply a BTS-internal ECU to the uplink traffic frame stream\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->use_ul_ecu = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_no_rtp_int_ul_ecu,
+      cfg_bts_no_rtp_int_ul_ecu_cmd,
+      "no rtp internal-uplink-ecu",
+      NO_STR RTP_STR "Apply a BTS-internal ECU to the uplink traffic frame stream\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->use_ul_ecu = false;
+	return CMD_SUCCESS;
+}
+
+DEFUN_ATTR(cfg_bts_rtp_hr_format,
+	   cfg_bts_rtp_hr_format_cmd,
+	   "rtp hr-format (rfc5993|ts101318)",
+	   RTP_STR "HRv1 codec output format\n" "RFC 5993\n" "TS 101 318\n",
+	   CMD_ATTR_IMMEDIATE)
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->emit_hr_rfc5993 = !strcmp(argv[0], "rfc5993");
+	return CMD_SUCCESS;
+}
+
 #define PAG_STR "Paging related parameters\n"
 
 DEFUN_ATTR(cfg_bts_paging_queue_size,
@@ -846,7 +1020,7 @@ DEFUN_ATTR(cfg_bts_max_ber_rach, cfg_bts_max_ber_rach_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_bts_pcu_sock, cfg_bts_pcu_sock_cmd,
+DEFUN(cfg_bts_pcu_sock_path, cfg_bts_pcu_sock_path_cmd,
 	"pcu-socket PATH",
 	"Configure the PCU socket file/path name\n"
 	"UNIX socket path\n")
@@ -859,6 +1033,16 @@ DEFUN(cfg_bts_pcu_sock, cfg_bts_pcu_sock_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_pcu_sock_ql, cfg_bts_pcu_sock_ql_cmd,
+	"pcu-socket-wqueue-length <1-" OSMO_STRINGIFY_VAL(BTS_CFG_PCU_SOCK_WQUEUE_LEN_MAX_MAX) ">",
+	"Configure the PCU socket queue length\n"
+	"Queue length\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->pcu.sock_wqueue_len_max = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
 DEFUN_ATTR(cfg_bts_supp_meas_toa256, cfg_bts_supp_meas_toa256_cmd,
 	   "supp-meas-info toa256",
 	   "Configure the RSL Supplementary Measurement Info\n"
@@ -1089,10 +1273,10 @@ static void bts_dump_vty_features(struct vty *vty, const struct gsm_bts *bts)
 
 	vty_out(vty, "  BTS model specific (internal) flags:%s", VTY_NEWLINE);
 
-	for (i = 0, no_features = true; i < sizeof(bts->flags) * 8; i++) {
-		if (bts_internal_flag_get(bts, (1 << i))) {
+	for (i = 0, no_features = true; i < _BTS_INTERNAL_FLAG_NUM; i++) {
+		if (bts_internal_flag_get(bts, i)) {
 			vty_out(vty, "    %03u ", i);
-			vty_out(vty, "%-40s%s", get_value_string(bts_impl_flag_desc, (1 << i)), VTY_NEWLINE);
+			vty_out(vty, "%-40s%s", get_value_string(bts_impl_flag_desc, i), VTY_NEWLINE);
 			no_features = false;
 		}
 	}
@@ -1131,7 +1315,7 @@ static void bts_dump_vty(struct vty *vty, const struct gsm_bts *bts)
 	vty_out(vty, "  NM State: ");
 	net_dump_nmstate(vty, &bts->mo.nm_state);
 	vty_out(vty, "  Site Mgr NM State: ");
-	net_dump_nmstate(vty, &bts->site_mgr.mo.nm_state);
+	net_dump_nmstate(vty, &g_bts_sm->mo.nm_state);
 	if (strnlen(bts->pcu_version, MAX_VERSION_LENGTH))
 		vty_out(vty, "  PCU version %s connected%s",
 			bts->pcu_version, VTY_NEWLINE);
@@ -1190,23 +1374,22 @@ DEFUN(show_bts, show_bts_cmd, "show bts [<0-255>]",
       SHOW_STR "Display information about a BTS\n"
       BTS_NR_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	int bts_nr;
 
 	if (argc != 0) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS '%s'%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts_dump_vty(vty, gsm_bts_num(net, bts_nr));
+		bts_dump_vty(vty, gsm_bts_num(g_bts_sm, bts_nr));
 		return CMD_SUCCESS;
 	}
 	/* print all BTS's */
-	for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++)
-		bts_dump_vty(vty, gsm_bts_num(net, bts_nr));
+	for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++)
+		bts_dump_vty(vty, gsm_bts_num(g_bts_sm, bts_nr));
 
 	return CMD_SUCCESS;
 }
@@ -1214,21 +1397,22 @@ DEFUN(show_bts, show_bts_cmd, "show bts [<0-255>]",
 static void gprs_dump_vty(struct vty *vty, const struct gsm_bts *bts)
 {
 	unsigned int i;
+	const struct gsm_gprs_nse *nse = &bts->site_mgr->gprs.nse;
 
 	/* GPRS parameters received from the BSC */
 	vty_out(vty, "BTS %u, RAC %u, NSEI %u, BVCI %u%s",
 		bts->nr, bts->gprs.rac,
-		bts->gprs.nse.nsei,
+		nse->nsei,
 		bts->gprs.cell.bvci,
 		VTY_NEWLINE);
 
 	vty_out(vty, "  Cell NM state: ");
 	net_dump_nmstate(vty, &bts->gprs.cell.mo.nm_state);
 	vty_out(vty, "  NSE NM state: ");
-	net_dump_nmstate(vty, &bts->gprs.nse.mo.nm_state);
+	net_dump_nmstate(vty, &nse->mo.nm_state);
 
-	for (i = 0; i < ARRAY_SIZE(bts->gprs.nsvc); i++) {
-		const struct gsm_bts_gprs_nsvc *nsvc = &bts->gprs.nsvc[i];
+	for (i = 0; i < ARRAY_SIZE(nse->nsvc); i++) {
+		const struct gsm_gprs_nsvc *nsvc = &nse->nsvc[i];
 
 		vty_out(vty, "  NSVC%u (NSVCI %u) NM state: ", i, nsvc->nsvci);
 		net_dump_nmstate(vty, &nsvc->mo.nm_state);
@@ -1263,10 +1447,9 @@ DEFUN(show_bts_gprs, show_bts_gprs_cmd,
       SHOW_STR "Display information about a BTS\n"
       BTS_NR_STR "GPRS/EGPRS configuration\n")
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts;
 
-	bts = gsm_bts_num(net, atoi(argv[0]));
+	bts = gsm_bts_num(g_bts_sm, atoi(argv[0]));
 	if (bts == NULL) {
 		vty_out(vty, "%% can't find BTS '%s'%s",
 			argv[0], VTY_NEWLINE);
@@ -1282,16 +1465,15 @@ DEFUN(test_send_failure_event_report, test_send_failure_event_report_cmd, "test
       "Various testing commands\n"
       "Send a test OML failure event report to the BSC\n" BTS_NR_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	int bts_nr = atoi(argv[0]);
 	const struct gsm_bts *bts;
 
-	if (bts_nr >= net->num_bts) {
+	if (bts_nr >= g_bts_sm->num_bts) {
 		vty_out(vty, "%% can't find BTS '%s'%s", argv[0], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
-	bts = gsm_bts_num(net, bts_nr);
+	bts = gsm_bts_num(g_bts_sm, bts_nr);
 	oml_tx_failure_event_rep(&bts->mo, NM_SEVER_MINOR, OSMO_EVT_WARN_SW_WARN, "test message sent from VTY");
 
 	return CMD_SUCCESS;
@@ -1303,9 +1485,8 @@ DEFUN_HIDDEN(radio_link_timeout, radio_link_timeout_cmd, "bts <0-0> radio-link-t
 		"Use infinite timeout (DANGEROUS: only use during testing!)\n"
 		"Number of lost SACCH blocks\n")
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	int bts_nr = atoi(argv[0]);
-	struct gsm_bts *bts = gsm_bts_num(net, bts_nr);
+	struct gsm_bts *bts = gsm_bts_num(g_bts_sm, bts_nr);
 
 	if (!bts) {
 		vty_out(vty, "%% can't find BTS '%s'%s", argv[0], VTY_NEWLINE);
@@ -1334,12 +1515,11 @@ DEFUN(bts_c0_power_red,
       "BCCH carrier power reduction operation\n"
       "Power reduction value (in dB, even numbers only)\n")
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	const int bts_nr = atoi(argv[0]);
 	const int red = atoi(argv[1]);
 	struct gsm_bts *bts;
 
-	bts = gsm_bts_num(net, atoi(argv[0]));
+	bts = gsm_bts_num(g_bts_sm, atoi(argv[0]));
 	if (bts == NULL) {
 		vty_out(vty, "%% No such BTS (%d)%s", bts_nr, VTY_NEWLINE);
 		return CMD_WARNING;
@@ -1470,10 +1650,10 @@ static void trx_dump_vty(struct vty *vty, const struct gsm_bts_trx *trx)
 
 	vty_out(vty, "  NM State: ");
 	net_dump_nmstate(vty, &trx->mo.nm_state);
-	vty_out(vty, "  RSL State: %s%s", trx->rsl_link? "connected" : "disconnected", VTY_NEWLINE);
+	vty_out(vty, "  RSL State: %s%s", trx->bb_transc.rsl.link ? "connected" : "disconnected", VTY_NEWLINE);
 	vty_out(vty, "  Baseband Transceiver NM State: ");
 	net_dump_nmstate(vty, &trx->bb_transc.mo.nm_state);
-	vty_out(vty, "  IPA stream ID: 0x%02x%s", trx->rsl_tei, VTY_NEWLINE);
+	vty_out(vty, "  IPA stream ID: 0x%02x%s", trx->bb_transc.rsl.tei, VTY_NEWLINE);
 }
 
 static inline void print_all_trx(struct vty *vty, const struct gsm_bts *bts)
@@ -1489,19 +1669,18 @@ DEFUN(show_trx,
 	SHOW_STR "Display information about a TRX\n"
 	BTS_TRX_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts = NULL;
 	int bts_nr, trx_nr;
 
 	if (argc >= 1) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS '%s'%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts = gsm_bts_num(net, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 	}
 	if (argc >= 2) {
 		trx_nr = atoi(argv[1]);
@@ -1519,8 +1698,8 @@ DEFUN(show_trx,
 		return CMD_SUCCESS;
 	}
 
-	for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++)
-		print_all_trx(vty, gsm_bts_num(net, bts_nr));
+	for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++)
+		print_all_trx(vty, gsm_bts_num(g_bts_sm, bts_nr));
 
 	return CMD_SUCCESS;
 }
@@ -1545,7 +1724,6 @@ DEFUN(show_ts,
 	SHOW_STR "Display information about a TS\n"
 	BTS_TRX_TS_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts = NULL;
 	const struct gsm_bts_trx *trx = NULL;
 	const struct gsm_bts_trx_ts *ts = NULL;
@@ -1554,12 +1732,12 @@ DEFUN(show_ts,
 	if (argc >= 1) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS '%s'%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts = gsm_bts_num(net, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 	}
 	if (argc >= 2) {
 		trx_nr = atoi(argv[1]);
@@ -1600,8 +1778,8 @@ DEFUN(show_ts,
 		}
 	} else {
 		/* Iterate over all BTS, TRX in each BTS, TS in each TRX */
-		for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++) {
-			bts = gsm_bts_num(net, bts_nr);
+		for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++) {
+			bts = gsm_bts_num(g_bts_sm, bts_nr);
 			for (trx_nr = 0; trx_nr < bts->num_trx; trx_nr++) {
 				trx = gsm_bts_trx_num(bts, trx_nr);
 				for (ts_nr = 0; ts_nr < TRX_NR_TS; ts_nr++) {
@@ -1689,10 +1867,10 @@ static void lchan_ms_power_ctrl_state_dump(struct vty *vty, unsigned int indent,
 	int current_dbm = ms_pwr_dbm(trx->bts->band, st->current);
 	int max_dbm = ms_pwr_dbm(trx->bts->band, st->max);
 
-	cfg_out(vty, "Current power level: %u, -%d dBm",
+	cfg_out(vty, "Current power level: %u, %d dBm",
 		st->current, current_dbm);
 	if (st->dpc_params != NULL)
-		vty_out(vty, " (max %u, -%d dBm)", st->max, max_dbm);
+		vty_out(vty, " (max %u, %d dBm)", st->max, max_dbm);
 	vty_out(vty, "%s", VTY_NEWLINE);
 
 	if (st->dpc_params == NULL)
@@ -1819,20 +1997,44 @@ static void lchan_dump_full_vty(struct vty *vty, const struct gsm_lchan *lchan)
 	vty_out(vty, "  Channel Mode / Codec: %s%s",
 		gsm48_chan_mode_name(lchan->tch_mode),
 		VTY_NEWLINE);
+	if (lchan->rsl_cmode == RSL_CMOD_SPD_DATA) {
+		vty_out(vty, "  CSD mode: %s%s",
+			lchan_csd_mode_desc(lchan->csd_mode),
+			VTY_NEWLINE);
+	}
+	if (lchan->tch_mode == GSM48_CMODE_SPEECH_AMR) {
+		const struct amr_multirate_conf *amr_mrc = &lchan->tch.amr_mr;
+		const struct gsm48_multi_rate_conf *mr_conf =
+			(const struct gsm48_multi_rate_conf *) amr_mrc->gsm48_ie;
+		vty_out(vty, "  AMR Multi-Rate Configuration: ICMI=%u, Start Mode=%u gsm48=%02x%02x%s",
+			mr_conf->icmi, mr_conf->smod, amr_mrc->gsm48_ie[0], amr_mrc->gsm48_ie[1], VTY_NEWLINE);
+		for (uint8_t i = 0; i < amr_mrc->num_modes; i++) {
+			const struct amr_mode *amode = &amr_mrc->mode[i];
+			vty_out(vty, "    AMR Mode %u (%s), threshold=%u, hysteresis=%u%s",
+				amode->mode, osmo_amr_type_name(amode->mode),
+				amode->threshold, amode->hysteresis, VTY_NEWLINE);
+		}
+	}
 
 	if (lchan->abis_ip.bound_ip) {
 		ia.s_addr = htonl(lchan->abis_ip.bound_ip);
-		vty_out(vty, "  Bound IP: %s Port %u RTP_TYPE2=%u CONN_ID=%u%s",
+		vty_out(vty, "  Bound IP: %s Port %u CONN_ID=%u",
 			inet_ntoa(ia), lchan->abis_ip.bound_port,
-			lchan->abis_ip.rtp_payload2, lchan->abis_ip.conn_id,
-			VTY_NEWLINE);
+			lchan->abis_ip.conn_id);
+		if (lchan->abis_ip.osmux.use)
+			vty_out(vty, " Osmux_CID=%u%s", lchan->abis_ip.osmux.local_cid, VTY_NEWLINE);
+		else
+			vty_out(vty, " RTP_TYPE2=%u%s", lchan->abis_ip.rtp_payload2, VTY_NEWLINE);
 	}
 	if (lchan->abis_ip.connect_ip) {
 		ia.s_addr = htonl(lchan->abis_ip.connect_ip);
-		vty_out(vty, "  Conn. IP: %s Port %u RTP_TYPE=%u SPEECH_MODE=0x%02u%s",
+		vty_out(vty, "  Conn. IP: %s Port %u SPEECH_MODE=0x%02x",
 			inet_ntoa(ia), lchan->abis_ip.connect_port,
-			lchan->abis_ip.rtp_payload, lchan->abis_ip.speech_mode,
-			VTY_NEWLINE);
+			lchan->abis_ip.speech_mode);
+		if (lchan->abis_ip.osmux.use)
+			vty_out(vty, " Osmux_CID=%u%s", lchan->abis_ip.osmux.remote_cid, VTY_NEWLINE);
+		else
+			vty_out(vty, " RTP_TYPE=%u%s", lchan->abis_ip.rtp_payload, VTY_NEWLINE);
 	}
 #define LAPDM_ESTABLISHED(link, sapi_idx) \
 		(link).datalink[sapi_idx].dl.state == LAPD_STATE_MF_EST
@@ -1945,7 +2147,6 @@ static int dump_lchan_bts(const struct gsm_bts *bts, struct vty *vty,
 static int lchan_summary(struct vty *vty, int argc, const char **argv,
 			 void (*dump_cb)(struct vty *, const struct gsm_lchan *))
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
 	const struct gsm_bts_trx *trx = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
 	const struct gsm_bts_trx_ts *ts = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
@@ -1955,12 +2156,12 @@ static int lchan_summary(struct vty *vty, int argc, const char **argv,
 	if (argc >= 1) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS %s%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts = gsm_bts_num(net, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 
 		if (argc == 1)
 			return dump_lchan_bts(bts, vty, dump_cb);
@@ -2001,8 +2202,8 @@ static int lchan_summary(struct vty *vty, int argc, const char **argv,
 		return CMD_SUCCESS;
 	}
 
-	for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++) {
-		bts = gsm_bts_num(net, bts_nr);
+	for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++) {
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 		dump_lchan_bts(bts, vty, dump_cb);
 	}
 
@@ -2028,8 +2229,7 @@ DEFUN(show_lchan_summary,
 	return lchan_summary(vty, argc, argv, lchan_dump_short_vty);
 }
 
-static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
-				       const char **argv, int idx)
+static struct gsm_lchan *resolve_lchan(const char **argv, int idx)
 {
 	int bts_nr = atoi(argv[idx+0]);
 	int trx_nr = atoi(argv[idx+1]);
@@ -2040,7 +2240,7 @@ static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
 	struct gsm_bts_trx *trx;
 	struct gsm_bts_trx_ts *ts;
 
-	bts = gsm_bts_num(net, bts_nr);
+	bts = gsm_bts_num(g_bts_sm, bts_nr);
 	if (!bts)
 		return NULL;
 
@@ -2093,6 +2293,22 @@ DEFUN(cfg_bts_gsmtap_remote_host,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_gsmtap_local_host,
+      cfg_bts_gsmtap_local_host_cmd,
+      "gsmtap-local-host HOSTNAME",
+      "Enable local bind for GSMTAP Um logging (see also 'gsmtap-sapi')\n"
+      "Local IP address or hostname\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	osmo_talloc_replace_string(bts, &bts->gsmtap.local_host, argv[0]);
+
+	if (vty->type != VTY_FILE)
+		vty_out(vty, "%% This command requires restart%s", VTY_NEWLINE);
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_bts_no_gsmtap_remote_host,
       cfg_bts_no_gsmtap_remote_host_cmd,
       "no gsmtap-remote-host",
@@ -2110,6 +2326,24 @@ DEFUN(cfg_bts_no_gsmtap_remote_host,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_no_gsmtap_local_host,
+      cfg_bts_no_gsmtap_local_host_cmd,
+      "no gsmtap-local-host",
+      NO_STR "Disable local bind for GSMTAP Um logging\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	if (bts->gsmtap.local_host != NULL)
+		talloc_free(bts->gsmtap.local_host);
+
+	bts->gsmtap.local_host = NULL;
+
+	if (vty->type != VTY_FILE)
+		vty_out(vty, "%% This command requires restart%s", VTY_NEWLINE);
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_bts_gsmtap_sapi_all, cfg_bts_gsmtap_sapi_all_cmd,
 	"gsmtap-sapi (enable-all|disable-all)",
 	"Enable/disable sending of UL/DL messages over GSMTAP\n"
@@ -2146,7 +2380,7 @@ DEFUN(cfg_bts_gsmtap_sapi, cfg_bts_gsmtap_sapi_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_trx_no_gsmtap_sapi, cfg_bts_no_gsmtap_sapi_cmd,
+DEFUN(cfg_bts_no_gsmtap_sapi, cfg_bts_no_gsmtap_sapi_cmd,
 	"HIDDEN", "HIDDEN")
 {
 	struct gsm_bts *bts = vty->index;
@@ -2163,6 +2397,29 @@ DEFUN(cfg_trx_no_gsmtap_sapi, cfg_bts_no_gsmtap_sapi_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_gsmtap_rlp, cfg_bts_gsmtap_rlp_cmd,
+	"gsmtap-rlp [skip-null]",
+	"Enable generation of GSMTAP frames for RLP (non-transparent CSD)\n"
+	"Skip the generation of GSMTAP for RLP NULL frames\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->gsmtap.rlp = true;
+	if (argc >= 1 && !strcmp(argv[0], "skip-null"))
+		bts->gsmtap.rlp_skip_null = true;
+	else
+		bts->gsmtap.rlp_skip_null = false;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_no_gsmtap_rlp, cfg_bts_no_gsmtap_rlp_cmd,
+	"no gsmtap-rlp",
+	NO_STR "Disable generation of GSMTAP frames for RLP (non-transparent CSD)\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->gsmtap.rlp = false;
+	return CMD_SUCCESS;
+}
+
 static struct cmd_node phy_node = {
 	PHY_NODE,
 	"%s(phy)# ",
@@ -2271,11 +2528,10 @@ DEFUN(bts_t_t_l_jitter_buf,
 	BTS_T_T_L_STR "RTP settings\n"
 	"Jitter buffer\n" "Size of jitter buffer in (ms)\n")
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct gsm_lchan *lchan;
 	int jitbuf_ms = atoi(argv[4]), rc;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2304,10 +2560,9 @@ DEFUN_ATTR(bts_t_t_l_loopback,
 	   BTS_T_T_L_STR "Set loopback\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2323,10 +2578,9 @@ DEFUN_ATTR(no_bts_t_t_l_loopback,
 	   NO_STR BTS_T_T_L_STR "Set loopback\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2349,13 +2603,12 @@ DEFUN_ATTR(bts_t_t_l_power_ctrl_mode,
 	   "Enable the power control loop\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_power_ctrl_params *params;
 	struct lchan_power_ctrl_state *state;
 	const char **args = argv + 4;
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2386,12 +2639,11 @@ DEFUN_ATTR(bts_t_t_l_power_ctrl_current_max,
 	   "BS power reduction (in dB) or MS power level\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct lchan_power_ctrl_state *state;
 	const char **args = argv + 4;
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2505,6 +2757,11 @@ int bts_vty_init(void *ctx)
 	install_element(BTS_NODE, &cfg_bts_rtp_port_range_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_ip_dscp_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_priority_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_cont_stream_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_rtp_cont_stream_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_int_ul_ecu_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_rtp_int_ul_ecu_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_hr_format_cmd);
 	install_element(BTS_NODE, &cfg_bts_band_cmd);
 	install_element(BTS_NODE, &cfg_description_cmd);
 	install_element(BTS_NODE, &cfg_no_description_cmd);
@@ -2519,7 +2776,8 @@ int bts_vty_init(void *ctx)
 	install_element(BTS_NODE, &cfg_bts_min_qual_rach_cmd);
 	install_element(BTS_NODE, &cfg_bts_min_qual_norm_cmd);
 	install_element(BTS_NODE, &cfg_bts_max_ber_rach_cmd);
-	install_element(BTS_NODE, &cfg_bts_pcu_sock_cmd);
+	install_element(BTS_NODE, &cfg_bts_pcu_sock_path_cmd);
+	install_element(BTS_NODE, &cfg_bts_pcu_sock_ql_cmd);
 	install_element(BTS_NODE, &cfg_bts_supp_meas_toa256_cmd);
 	install_element(BTS_NODE, &cfg_bts_no_supp_meas_toa256_cmd);
 	install_element(BTS_NODE, &cfg_bts_smscb_max_qlen_cmd);
@@ -2528,9 +2786,24 @@ int bts_vty_init(void *ctx)
 
 	install_element(BTS_NODE, &cfg_bts_gsmtap_remote_host_cmd);
 	install_element(BTS_NODE, &cfg_bts_no_gsmtap_remote_host_cmd);
+	install_element(BTS_NODE, &cfg_bts_gsmtap_local_host_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_gsmtap_local_host_cmd);
 	install_element(BTS_NODE, &cfg_bts_gsmtap_sapi_all_cmd);
 	install_element(BTS_NODE, &cfg_bts_gsmtap_sapi_cmd);
 	install_element(BTS_NODE, &cfg_bts_no_gsmtap_sapi_cmd);
+	install_element(BTS_NODE, &cfg_bts_gsmtap_rlp_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_gsmtap_rlp_cmd);
+
+	/* Osmux Node */
+	install_element(BTS_NODE, &cfg_bts_osmux_cmd);
+	install_node(&osmux_node, config_write_dummy);
+
+	install_element(OSMUX_NODE, &cfg_bts_osmux_use_cmd);
+	install_element(OSMUX_NODE, &cfg_bts_osmux_ip_cmd);
+	install_element(OSMUX_NODE, &cfg_bts_osmux_port_cmd);
+	install_element(OSMUX_NODE, &cfg_bts_osmux_batch_factor_cmd);
+	install_element(OSMUX_NODE, &cfg_bts_osmux_batch_size_cmd);
+	install_element(OSMUX_NODE, &cfg_bts_osmux_dummy_padding_cmd);
 
 	/* add and link to TRX config node */
 	install_element(BTS_NODE, &cfg_bts_trx_cmd);
