@@ -146,12 +146,12 @@ int l1if_tch_rx(struct gsm_bts_trx *trx, uint8_t chan_nr,
 	    &trx->ts[L1SAP_CHAN2TS(chan_nr)].lchan[l1sap_chan2ss(chan_nr)];
 
 	if (data_ind->Data.ulDataLength < 1) {
-		LOGPFN(DL1P, LOGL_DEBUG, fn, "chan_nr %d Rx Payload size 0\n", chan_nr);
+		LOGPLCFN(lchan, fn, DL1P, LOGL_DEBUG, "chan_nr %d Rx Payload size 0\n", chan_nr);
 		/* Push empty payload to upper layers */
 		rmsg = msgb_alloc_headroom(256, 128, "L1P-to-RTP");
 		return add_l1sap_header(trx, rmsg, lchan, chan_nr,
 					data_ind->Data.ulFrameNumber,
-					ber10k, lqual_cb);
+					ber10k, lqual_cb, 0, 0, 0);
 	}
 
 	payload_len = data_ind->Data.ulDataLength;
@@ -173,13 +173,13 @@ int l1if_tch_rx(struct gsm_bts_trx *trx, uint8_t chan_nr,
 			goto err_payload_match;
 		break;
 	default:
-		LOGPFN(DL1P, LOGL_NOTICE, fn, "%s Rx Payload Type %d is unsupported\n",
-		     gsm_lchan_name(lchan), payload_type);
+		LOGPLCFN(lchan, fn, DL1P, LOGL_NOTICE, "%s Rx Payload Type %d is unsupported\n",
+			 gsm_lchan_name(lchan), payload_type);
 		break;
 	}
 
-	LOGPFN(DL1P, LOGL_DEBUG, fn, "%s Rx codec frame (%u): %s\n", gsm_lchan_name(lchan),
-		payload_len, osmo_hexdump(payload, payload_len));
+	LOGPLCFN(lchan, fn, DL1P, LOGL_DEBUG, "%s Rx codec frame (%u): %s\n", gsm_lchan_name(lchan), payload_len,
+		 osmo_hexdump(payload, payload_len));
 
 	switch (payload_type) {
 	case cOCTVC1_GSM_PAYLOAD_TYPE_ENUM_FULL_RATE:
@@ -201,7 +201,7 @@ int l1if_tch_rx(struct gsm_bts_trx *trx, uint8_t chan_nr,
 		rmsg = l1_to_rtppayload_amr(payload, payload_len,
 				&lchan->tch.amr_mr);
 #else
-		LOGPFN(DL1P, LOGL_ERROR, fn, "OctPHY only supports FR!\n");
+		LOGPLCFN(lchan, fn, DL1P, LOGL_ERROR, "OctPHY only supports FR!\n");
 		return -1;
 #endif
 		break;
@@ -210,13 +210,13 @@ int l1if_tch_rx(struct gsm_bts_trx *trx, uint8_t chan_nr,
 	if (rmsg)
 		return add_l1sap_header(trx, rmsg, lchan, chan_nr,
 					data_ind->Data.ulFrameNumber,
-					ber10k, lqual_cb);
+					ber10k, lqual_cb, 0, 0, 0);
 
 	return 0;
 
 err_payload_match:
-	LOGPFN(DL1P, LOGL_ERROR, fn, "%s Rx Payload Type %d incompatible with lchan\n",
-	     gsm_lchan_name(lchan), payload_type);
+	LOGPLCFN(lchan, fn, DL1P, LOGL_ERROR, "%s Rx Payload Type %d incompatible with lchan\n",
+		 gsm_lchan_name(lchan), payload_type);
 	return -EINVAL;
 }
 
@@ -241,8 +241,7 @@ void l1if_tch_encode(struct gsm_lchan *lchan, uint32_t *payload_type,
 	uint8_t *l1_payload;
 	int rc = -1;
 
-	DEBUGP(DRTP, "%s RTP IN: %s\n", gsm_lchan_name(lchan),
-	       osmo_hexdump(rtp_pl, rtp_pl_len));
+	LOGPLCHAN(lchan, DRTP, LOGL_DEBUG, "RTP IN: %s\n", osmo_hexdump(rtp_pl, rtp_pl_len));
 
 	l1_payload = &data[0];
 
@@ -271,13 +270,11 @@ void l1if_tch_encode(struct gsm_lchan *lchan, uint32_t *payload_type,
 	}
 
 	if (rc < 0) {
-		LOGP(DRTP, LOGL_ERROR, "%s unable to parse RTP payload\n",
-		     gsm_lchan_name(lchan));
+		LOGPLCHAN(lchan, DRTP, LOGL_ERROR, "unable to parse RTP payload\n");
 		return;
 	}
 
 	*len = rc;
 
-	DEBUGP(DRTP, "%s RTP->L1: %s\n", gsm_lchan_name(lchan),
-	       osmo_hexdump(data, *len));
+	LOGPLCHAN(lchan, DRTP, LOGL_DEBUG, "RTP->L1: %s\n", osmo_hexdump(data, *len));
 }
