@@ -13,7 +13,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -132,6 +132,8 @@ int bts_model_init(struct gsm_bts *bts)
 	bts->model_priv = bts_trx;
 	bts->variant = BTS_OSMO_TRX;
 	bts->support.ciphers = CIPHER_A5(1) | CIPHER_A5(2) | CIPHER_A5(3) | CIPHER_A5(4);
+	bts->gprs.cell.support.gprs_codings = NM_IPAC_MASK_GPRS_CODING_CS
+					    | NM_IPAC_MASK_GPRS_CODING_MCS;
 
 	/* The nominal value for each TRX is later overwritten through VTY cmd
 	 * 'nominal-tx-power' if present, otherwise through TRXC cmd NOMTXPOWER.
@@ -154,15 +156,47 @@ int bts_model_init(struct gsm_bts *bts)
 	osmo_bts_set_feature(bts->features, BTS_FEAT_SPEECH_H_AMR);
 	osmo_bts_set_feature(bts->features, BTS_FEAT_SPEECH_H_V1);
 	osmo_bts_set_feature(bts->features, BTS_FEAT_VAMOS);
+	osmo_bts_set_feature(bts->features, BTS_FEAT_VGCS);
+	osmo_bts_set_feature(bts->features, BTS_FEAT_VBS);
 
 	bts_internal_flag_set(bts, BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB);
 	bts_internal_flag_set(bts, BTS_INTERNAL_FLAG_INTERF_MEAS);
 
+	/* The default HR codec output format in the absence of saved
+	 * vty config needs to match what was implemented previously,
+	 * for the sake of existing deployments, i.e., to avoid
+	 * a surprise functional change upon software update. */
+	bts->emit_hr_rfc5993 = true;
+
+	/* For the same reason as the above, rtp internal-uplink-ecu
+	 * needs to be enabled by default on osmo-bts-trx model only. */
+	bts->use_ul_ecu = true;
+
 	return 0;
 }
 
 int bts_model_trx_init(struct gsm_bts_trx *trx)
 {
+	/* Frequency bands indicated to the BSC */
+	trx->support.freq_bands = NM_IPAC_F_FREQ_BAND_PGSM
+				| NM_IPAC_F_FREQ_BAND_EGSM
+				| NM_IPAC_F_FREQ_BAND_RGSM
+				| NM_IPAC_F_FREQ_BAND_DCS
+				| NM_IPAC_F_FREQ_BAND_PCS
+				| NM_IPAC_F_FREQ_BAND_850
+				| NM_IPAC_F_FREQ_BAND_480
+				| NM_IPAC_F_FREQ_BAND_450;
+
+	/* Channel types and modes indicated to the BSC */
+	trx->support.chan_types = NM_IPAC_MASK_CHANT_COMMON
+				| NM_IPAC_F_CHANT_BCCH_SDCCH4_CBCH
+				| NM_IPAC_F_CHANT_SDCCH8_CBCH
+				| NM_IPAC_F_CHANT_PDCHF
+				| NM_IPAC_F_CHANT_TCHF_PDCHF;
+	trx->support.chan_modes = NM_IPAC_MASK_CHANM_SPEECH
+				| NM_IPAC_MASK_CHANM_CSD_NT
+				| NM_IPAC_MASK_CHANM_CSD_T;
+
 	/* The nominal value for each TRX is later overwritten through VTY cmd
 	 * 'nominal-tx-power' if present, otherwise through TRXC cmd NOMTXPOWER.
 	 */
