@@ -42,6 +42,7 @@
 #include <osmo-bts/amr.h>
 #include <osmo-bts/abis.h>
 #include <osmo-bts/scheduler.h>
+#include <osmo-bts/handover.h>
 #include "virtual_um.h"
 
 extern int vbts_sched_start(struct gsm_bts *bts);
@@ -187,6 +188,11 @@ nomessage:
 /* called by common part once OML link is established */
 int bts_model_oml_estab(struct gsm_bts *bts)
 {
+	struct phy_instance *pinst = trx_phy_instance(bts->c0);
+
+	if (vbts_sched_start(pinst->trx->bts) < 0)
+		return -ENOLINK;
+
 	return 0;
 }
 
@@ -217,17 +223,6 @@ int bts_model_phy_link_open(struct phy_link *plink)
 		if (pinst->trx == NULL)
 			continue;
 		trx_sched_init(pinst->trx);
-		/* Only start the scheduler for the transceiver on C0.
-		 * If we have multiple transceivers, CCCH is always on C0
-		 * and has to be auto active */
-		/* Other TRX are activated via OML by a PRIM_INFO_MODIFY
-		 * / PRIM_INFO_ACTIVATE */
-		if (pinst->trx == pinst->trx->bts->c0) {
-			vbts_sched_start(pinst->trx->bts);
-			/* FIXME: This is probably the wrong location to set the CCCH to active... the OML link def. needs to be reworked and fixed. */
-			pinst->trx->ts[0].lchan[CCCH_LCHAN].rel_act_kind = LCHAN_REL_ACT_OML;
-			lchan_set_state(&pinst->trx->ts[0].lchan[CCCH_LCHAN], LCHAN_S_ACTIVE);
-		}
 	}
 
 	/* this will automatically update the MO state of all associated TRX objects */
@@ -317,10 +312,10 @@ static int l1if_process_meas_res(struct gsm_bts_trx *trx, uint8_t tn, uint32_t f
 	/* 100% BER is n_bits_total is 0 */
 	float ber = n_bits_total==0 ? 1.0 : (float)n_errors / (float)n_bits_total;
 
-	DEBUGPFN(DMEAS, fn, "RX L1 frame %s chan_nr=0x%02x MS pwr=%ddBm rssi=%.1f dBFS "
-		"ber=%.2f%% (%d/%d bits) L1_ta=%d ta_ctrl.current=%d toa=%.2f\n",
-		gsm_lchan_name(lchan), chan_nr, ms_pwr_dbm(lchan->ts->trx->bts->band, lchan->ms_power_ctrl.max),
-		rssi, ber*100, n_errors, n_bits_total, lchan->meas.l1_info.ta, lchan->ta_ctrl.current, toa);
+	LOGPLCFN(lchan, fn, DMEAS, LOGL_DEBUG, "RX L1 frame chan_nr=0x%02x MS pwr=%ddBm rssi=%.1f dBFS "
+		 "ber=%.2f%% (%d/%d bits) L1_ta=%d ta_ctrl.current=%d toa=%.2f\n", chan_nr,
+		 ms_pwr_dbm(lchan->ts->trx->bts->band, lchan->ms_power_ctrl.max), rssi, ber * 100, n_errors,
+		 n_bits_total, lchan->meas.l1_info.ta, lchan->ta_ctrl.current, toa);
 
 	l1if_fill_meas_res(&l1sap, chan_nr, lchan->ta_ctrl.current + toa, ber, rssi, fn);
 
@@ -334,7 +329,6 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 {
 	struct msgb *msg = l1sap->oph.msg;
 	uint8_t chan_nr;
-	uint8_t tn, ss;
 	int rc = 0;
 	struct gsm_lchan *lchan;
 
@@ -350,77 +344,78 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 		/* put data into scheduler's queue */
 		return trx_sched_tch_req(trx, l1sap);
 	case OSMO_PRIM(PRIM_MPH_INFO, PRIM_OP_REQUEST):
+		if (l1sap->u.info.type == PRIM_INFO_ACT_CIPH)
+			chan_nr = l1sap->u.info.u.ciph_req.chan_nr;
+		else /* u.act_req used by PRIM_INFO_{ACTIVATE,DEACTIVATE,MODIFY} */
+			chan_nr = l1sap->u.info.u.act_req.chan_nr;
+		lchan = get_lchan_by_chan_nr(trx, chan_nr);
+		if (OSMO_UNLIKELY(lchan == NULL)) {
+			LOGP(DL1C, LOGL_ERROR,
+			     "Rx MPH-INFO.req (type=0x%02x) for non-existent lchan (%s)\n",
+			     l1sap->u.info.type, rsl_chan_nr_str(chan_nr));
+			rc = -ENODEV;
+			break;
+		}
+
 		switch (l1sap->u.info.type) {
 		case PRIM_INFO_ACT_CIPH:
-			chan_nr = l1sap->u.info.u.ciph_req.chan_nr;
-			tn = L1SAP_CHAN2TS(chan_nr);
-			ss = l1sap_chan2ss(chan_nr);
-			lchan = &trx->ts[tn].lchan[ss];
 			if (l1sap->u.info.u.ciph_req.uplink)
 				l1if_set_ciphering(lchan, chan_nr, 0);
 			if (l1sap->u.info.u.ciph_req.downlink)
 				l1if_set_ciphering(lchan, chan_nr, 1);
 			break;
 		case PRIM_INFO_ACTIVATE:
-		case PRIM_INFO_DEACTIVATE:
-		case PRIM_INFO_MODIFY:
-			chan_nr = l1sap->u.info.u.act_req.chan_nr;
-			tn = L1SAP_CHAN2TS(chan_nr);
-			ss = l1sap_chan2ss(chan_nr);
-			lchan = &trx->ts[tn].lchan[ss];
-			/* we receive a channel activation request from the BSC,
-			 * e.g. as a response to a channel req on RACH */
-			if (l1sap->u.info.type == PRIM_INFO_ACTIVATE) {
-				if ((chan_nr & 0xE0) == 0x80) {
-					LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Cannot activate"
-						  " channel %s\n", rsl_chan_nr_str(chan_nr));
-					break;
-				}
-				/* activate dedicated channel */
-				trx_sched_set_lchan(lchan, chan_nr, LID_DEDIC, true);
-				/* activate associated channel */
-				trx_sched_set_lchan(lchan, chan_nr, LID_SACCH, true);
-				/* set mode */
-				trx_sched_set_mode(lchan->ts, chan_nr,
-					lchan->rsl_cmode, lchan->tch_mode,
-					lchan->tch.amr_mr.num_modes,
-					lchan->tch.amr_mr.mode[0].mode,
-					lchan->tch.amr_mr.mode[1].mode,
-					lchan->tch.amr_mr.mode[2].mode,
-					lchan->tch.amr_mr.mode[3].mode,
-					amr_get_initial_mode(lchan),
-					(lchan->ho.active == 1));
-				/* set lchan active */
-				lchan_set_state(lchan, LCHAN_S_ACTIVE);
-				/* set initial ciphering */
-				l1if_set_ciphering(lchan, chan_nr, 0);
-				l1if_set_ciphering(lchan, chan_nr, 1);
-				if (lchan->encr.alg_id)
-					lchan->ciph_state = LCHAN_CIPH_RXTX_CONF;
-				else
-					lchan->ciph_state = LCHAN_CIPH_NONE;
+			if ((chan_nr & 0xE0) == 0x80) {
+				LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Cannot activate"
+					  " channel %s\n", rsl_chan_nr_str(chan_nr));
+				rc = -EPERM;
+				break;
+			}
+			/* activate dedicated channel */
+			trx_sched_set_lchan(lchan, chan_nr, LID_DEDIC, true);
+			/* activate associated channel */
+			trx_sched_set_lchan(lchan, chan_nr, LID_SACCH, true);
+			/* set mode */
+			trx_sched_set_mode(lchan->ts, chan_nr,
+					   lchan->rsl_cmode, lchan->tch_mode,
+					   lchan->tch.amr_mr.num_modes,
+					   lchan->tch.amr_mr.mode[0].mode,
+					   lchan->tch.amr_mr.mode[1].mode,
+					   lchan->tch.amr_mr.mode[2].mode,
+					   lchan->tch.amr_mr.mode[3].mode,
+					   amr_get_initial_mode(lchan),
+					   (lchan->ho.active == HANDOVER_ENABLED ||
+					    rsl_chan_rt_is_asci(lchan->rsl_chan_rt)));
+			/* set lchan active */
+			lchan_set_state(lchan, LCHAN_S_ACTIVE);
+			/* set initial ciphering */
+			l1if_set_ciphering(lchan, chan_nr, 0);
+			l1if_set_ciphering(lchan, chan_nr, 1);
+			if (lchan->encr.alg_id)
+				lchan->ciph_state = LCHAN_CIPH_RXTX_CONF;
+			else
+				lchan->ciph_state = LCHAN_CIPH_NONE;
 
-				/* confirm */
-				mph_info_chan_confirm(trx, chan_nr,
-					PRIM_INFO_ACTIVATE, 0);
-				break;
-			}
-			if (l1sap->u.info.type == PRIM_INFO_MODIFY) {
-				/* change mode */
-				trx_sched_set_mode(lchan->ts, chan_nr,
-					lchan->rsl_cmode, lchan->tch_mode,
-					lchan->tch.amr_mr.num_modes,
-					lchan->tch.amr_mr.mode[0].mode,
-					lchan->tch.amr_mr.mode[1].mode,
-					lchan->tch.amr_mr.mode[2].mode,
-					lchan->tch.amr_mr.mode[3].mode,
-					amr_get_initial_mode(lchan),
-					0);
-				break;
-			}
+			/* confirm */
+			mph_info_chan_confirm(trx, chan_nr, PRIM_INFO_ACTIVATE, 0);
+			break;
+		case PRIM_INFO_MODIFY:
+			/* change mode */
+			trx_sched_set_mode(lchan->ts, chan_nr,
+					   lchan->rsl_cmode, lchan->tch_mode,
+					   lchan->tch.amr_mr.num_modes,
+					   lchan->tch.amr_mr.mode[0].mode,
+					   lchan->tch.amr_mr.mode[1].mode,
+					   lchan->tch.amr_mr.mode[2].mode,
+					   lchan->tch.amr_mr.mode[3].mode,
+					   amr_get_initial_mode(lchan),
+					   0);
+			break;
+		case PRIM_INFO_DEACTIVATE:
 			if ((chan_nr & 0xE0) == 0x80) {
 				LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Cannot deactivate"
 					  " channel %s\n", rsl_chan_nr_str(chan_nr));
+				rc = -EPERM;
 				break;
 			}
 			/* deactivate associated channel */
