@@ -12,7 +12,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
  * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
@@ -27,45 +27,31 @@
 #include <osmocom/core/talloc.h>
 #include "osmo_mcast_sock.h"
 #include "virtual_um.h"
+
 #include <unistd.h>
+#include <errno.h>
 
 /**
- * Virtual UM interface file descriptor callback.
- * Should be called by select.c when the fd is ready for reading.
+ * Virtual UM interface file descriptor read callback.
  */
-static int virt_um_fd_cb(struct osmo_fd *ofd, unsigned int what)
+static void virt_um_read_cb(int rc, struct msgb *msg, void *data)
 {
-	struct virt_um_inst *vui = ofd->data;
+	struct virt_um_inst *vui = data;
+	msg->l1h = msg->data;
 
-	if (what & BSC_FD_READ) {
-		struct msgb *msg = msgb_alloc(VIRT_UM_MSGB_SIZE, "Virtual UM Rx");
-		int rc;
-
-		/* read message from fd into message buffer */
-		rc = mcast_bidir_sock_rx(vui->mcast_sock, msgb_data(msg), msgb_tailroom(msg));
-		if (rc > 0) {
-			msgb_put(msg, rc);
-			msg->l1h = msgb_data(msg);
-			/* call the l1 callback function for a received msg */
-			vui->recv_cb(vui, msg);
-		} else if (rc == 0) {
-			vui->recv_cb(vui, NULL);
-			osmo_fd_close(ofd);
-		} else
-			perror("Read from multicast socket");
-
-	}
-
-	return 0;
+	/* call the l1 callback function for a received msg */
+	vui->recv_cb(vui, msg);
 }
 
 struct virt_um_inst *virt_um_init(void *ctx, char *tx_mcast_group, uint16_t tx_mcast_port,
-				  char *rx_mcast_group, uint16_t rx_mcast_port,
+				  char *rx_mcast_group, uint16_t rx_mcast_port, int ttl, const char *dev_name,
 				  void (*recv_cb)(struct virt_um_inst *vui, struct msgb *msg))
 {
 	struct virt_um_inst *vui = talloc_zero(ctx, struct virt_um_inst);
+	int rc;
+
 	vui->mcast_sock = mcast_bidir_sock_setup(ctx, tx_mcast_group, tx_mcast_port,
-						 rx_mcast_group, rx_mcast_port, 1, virt_um_fd_cb, vui);
+						 rx_mcast_group, rx_mcast_port, 1, virt_um_read_cb, vui);
 	if (!vui->mcast_sock) {
 		perror("Unable to create VirtualUm multicast socket");
 		talloc_free(vui);
@@ -73,8 +59,37 @@ struct virt_um_inst *virt_um_init(void *ctx, char *tx_mcast_group, uint16_t tx_m
 	}
 	vui->recv_cb = recv_cb;
 
+	/* -1 means default, i.e. no TTL explicitly configured in VTY */
+	if (ttl >= 0) {
+		int txfd = osmo_iofd_get_fd(vui->mcast_sock->tx_iofd);
+		rc = osmo_sock_mcast_ttl_set(txfd, ttl);
+		if (rc < 0) {
+			perror("Cannot set TTL of Virtual Um transmit socket");
+			goto out_close;
+		}
+	}
+
+	if (dev_name) {
+		int txfd = osmo_iofd_get_fd(vui->mcast_sock->tx_iofd);
+		rc = osmo_sock_mcast_iface_set(txfd, dev_name);
+		if (rc < 0) {
+			perror("Cannot bind multicast tx to given device");
+			goto out_close;
+		}
+		int rxfd = osmo_iofd_get_fd(vui->mcast_sock->rx_iofd);
+		rc = osmo_sock_mcast_iface_set(rxfd, dev_name);
+		if (rc < 0) {
+			perror("Cannot bind multicast rx to given device");
+			goto out_close;
+		}
+	}
+
 	return vui;
 
+out_close:
+	mcast_bidir_sock_close(vui->mcast_sock);
+	talloc_free(vui);
+	return NULL;
 }
 
 void virt_um_destroy(struct virt_um_inst *vui)
@@ -90,11 +105,11 @@ int virt_um_write_msg(struct virt_um_inst *vui, struct msgb *msg)
 {
 	int rc;
 
-	rc = mcast_bidir_sock_tx(vui->mcast_sock, msgb_data(msg),
-	                msgb_length(msg));
-	if (rc < 0)
-		perror("Writing to multicast socket");
-	msgb_free(msg);
+	rc = mcast_bidir_sock_tx_msg(vui->mcast_sock, msg);
+	if (rc < 0) {
+		msgb_free(msg);
+		rc = -errno;
+	}
 
 	return rc;
 }
