@@ -53,12 +53,24 @@ static const struct test_case tests[] = {
 		.tch_mode = GSM48_CMODE_DATA_12k0,
 		.csd_mode = LCHAN_CSD_M_T_9600,
 	},
+	{
+		.name = "TCH/F9.6 NT",
+		.lchan_type = GSM_LCHAN_TCH_F,
+		.tch_mode = GSM48_CMODE_DATA_12k0,
+		.csd_mode = LCHAN_CSD_M_NT,
+	},
 	{
 		.name = "TCH/F4.8",
 		.lchan_type = GSM_LCHAN_TCH_F,
 		.tch_mode = GSM48_CMODE_DATA_6k0,
 		.csd_mode = LCHAN_CSD_M_T_4800,
 	},
+	{
+		.name = "TCH/F4.8 NT",
+		.lchan_type = GSM_LCHAN_TCH_F,
+		.tch_mode = GSM48_CMODE_DATA_6k0,
+		.csd_mode = LCHAN_CSD_M_NT,
+	},
 	{
 		.name = "TCH/H4.8",
 		.lchan_type = GSM_LCHAN_TCH_H,
@@ -81,7 +93,7 @@ static const struct test_case tests[] = {
 
 static void exec_test_case(const struct test_case *tc)
 {
-	const struct csd_v110_frame_desc *desc;
+	const struct csd_v110_lchan_desc *desc;
 	uint8_t rtp[RFC4040_RTP_PLEN] = { 0 };
 	ubit_t data_enc[BBUF_MAX];
 	ubit_t data_dec[BBUF_MAX];
@@ -89,13 +101,10 @@ static void exec_test_case(const struct test_case *tc)
 
 	/* obtain a V.110 frame description for the given channel type/rate */
 	OSMO_ASSERT(tc->tch_mode < ARRAY_SIZE(csd_v110_lchan_desc));
-	if (tc->lchan_type == GSM_LCHAN_TCH_F)
-		desc = &csd_v110_lchan_desc[tc->tch_mode].fr;
-	else
-		desc = &csd_v110_lchan_desc[tc->tch_mode].hr;
+	desc = &csd_v110_lchan_desc[tc->tch_mode];
 
 	/* total number of bits carried by a radio interface block */
-	const unsigned int bit_num = desc->num_bits * desc->num_blocks;
+	const unsigned int bit_num = CSD_V110_NUM_BITS(desc);
 	if (bit_num == 0) {
 		fprintf(stderr, "[i] Skipping '%s' (not implemented)\n", tc->name);
 		return;
@@ -115,7 +124,7 @@ static void exec_test_case(const struct test_case *tc)
 		data_enc[i] = i & 0x01;
 
 	/* encode an RTP frame and print it */
-	rc = csd_v110_rtp_encode(&lchan, &rtp[0], &data_enc[0], bit_num);
+	rc = csd_v110_rtp_encode(&lchan, &rtp[0], &data_enc[0], bit_num, 0);
 	fprintf(stderr, "[i] csd_v110_rtp_encode() returns %d\n", rc);
 	if (rc != RFC4040_RTP_PLEN)
 		return;
@@ -124,7 +133,7 @@ static void exec_test_case(const struct test_case *tc)
 		fprintf(stderr, "    %s\n", osmo_hexdump(&rtp[i * 16], 16));
 
 	/* decode the encoded RTP frame */
-	rc = csd_v110_rtp_decode(&lchan, &data_dec[0], &rtp[0], sizeof(rtp));
+	rc = csd_v110_rtp_decode(&lchan, &data_dec[0], NULL, &rtp[0], sizeof(rtp));
 	fprintf(stderr, "[i] csd_v110_rtp_decode() returns %d\n", rc);
 	if (rc != bit_num)
 		return;
@@ -136,10 +145,14 @@ static void exec_test_case(const struct test_case *tc)
 			i, data_dec[i], data_enc[i]);
 	}
 
+	/* for TCH/F14.4, we always expect a valid block */
+	if (tc->tch_mode == GSM48_CMODE_DATA_14k5)
+		return;
+
 	fprintf(stderr, "[i] Testing '%s' (IDLE)\n", tc->name);
 
 	/* encode an idle RTP frame and print it */
-	rc = csd_v110_rtp_encode(&lchan, &rtp[0], &data_enc[0], 0);
+	rc = csd_v110_rtp_encode(&lchan, &rtp[0], &data_enc[0], 0, 0);
 	fprintf(stderr, "[i] csd_v110_rtp_encode() returns %d\n", rc);
 	if (rc != RFC4040_RTP_PLEN)
 		return;
