@@ -8,6 +8,7 @@
 #include <osmo-bts/gsm_data.h>
 #include <osmo-bts/logging.h>
 #include <osmo-bts/bts.h>
+#include <osmo-bts/bts_sm.h>
 #include <osmo-bts/measurement.h>
 #include <osmo-bts/rsl.h>
 
@@ -455,9 +456,20 @@ int main(int argc, char **argv)
 	msgb_talloc_ctx_init(tall_bts_ctx, 0);
 
 	osmo_init_logging2(tall_bts_ctx, &bts_log_info);
-	osmo_stderr_target->categories[DMEAS].loglevel = LOGL_DEBUG;
+	log_set_log_level(osmo_stderr_target, LOGL_DEBUG);
+	log_set_print_category_hex(osmo_stderr_target, 0);
+	log_set_print_category(osmo_stderr_target, 0);
+	log_set_print_level(osmo_stderr_target, 1);
+	log_set_print_filename2(osmo_stderr_target, LOG_FILENAME_NONE);
+	log_set_use_color(osmo_stderr_target, 0);
+	log_parse_category_mask(osmo_stderr_target, "DMEAS,1:");
 
-	bts = gsm_bts_alloc(tall_bts_ctx, 0);
+	g_bts_sm = gsm_bts_sm_alloc(tall_bts_ctx);
+	if (!g_bts_sm) {
+		fprintf(stderr, "Failed to create BTS Site Manager structure\n");
+		exit(1);
+	}
+	bts = gsm_bts_alloc(g_bts_sm, 0);
 	if (!bts) {
 		fprintf(stderr, "Failed to create BTS structure\n");
 		exit(1);
